# """Test scripts for BSDGO1 Outlets."""
# from __future__ import annotations
# from typing import TYPE_CHECKING
# import orjson
# from pyvesync.devices.vesyncoutlet import VeSyncOutletBSDGO1
# from pyvesync.utils.helpers import Helpers as Helpers
# from pyvesync.models.vesync_models import ResponseDeviceDetailsModel
# from pyvesync.device_map import get_outlet
# import call_json
# import call_json_outlets
# from base_test_cases import TestBase
# from defaults import Defaults

# if TYPE_CHECKING:
#     from pyvesync.device_map import OutletMap

# DEVICE_TYPE = 'BSDOG01'
# if get_outlet(DEVICE_TYPE) is not None:
#     DEVICE_MAP = get_outlet(DEVICE_TYPE)

# DEV_LIST_DETAIL = call_json.DeviceList.device_list_item(DEVICE_MAP)

# CORRECT_BSDGO1_DETAILS = call_json_outlets.DETAILS_RESPONSES[DEVICE_TYPE]

# BAD_BSDGO1_LIST = call_json.DETAILS_BADCODE


# class TestVeSyncBSDGO1Switch(TestBase):
#     """Test BSDGO1 outlet API."""

#     def device_map(self) -> OutletMap:
#         outlet_map = get_outlet(DEVICE_TYPE)
#         if outlet_map is None:
#             raise ValueError(f'No device map found for {DEVICE_TYPE}')
#         return outlet_map

#     def device_list_model(self) -> ResponseDeviceDetailsModel:
#         return ResponseDeviceDetailsModel.from_dict(call_json.DeviceList.device_list_item(self.device_map()))

#     def test_bsdgo1_conf(self):
#         """Test initialization of BSDGO1 outlet."""
#         self.mock_api.return_value = call_json.DeviceList.device_list_response(DEVICE_TYPE)
#         self.run_in_loop(self.manager.get_devices)
#         outlets = self.manager.devices.outlets
#         assert len(outlets) == 1
#         bsdgo1_outlet = outlets[0]
#         assert isinstance(bsdgo1_outlet, VeSyncOutletBSDGO1)
#         assert bsdgo1_outlet.device_name == call_json.Defaults.name(DEVICE_TYPE)
#         assert bsdgo1_outlet.device_type == DEVICE_TYPE
#         assert bsdgo1_outlet.cid == call_json.Defaults.cid(DEVICE_TYPE)
#         assert bsdgo1_outlet.uuid == call_json.Defaults.uuid(DEVICE_TYPE)

#     def test_bsdgo1_details(self):
#         """Test BSDGO1 get_details()."""
#         resp_dict, status = self.device_list_model(), 200
#         self.mock_api.return_value = orjson.dumps(resp_dict), status

#         bsdgo1_outlet = VeSyncOutletBSDGO1(
#             self.device_list_model(),
#             self.manager,
#             self.device_map(),
#         )
#         self.run_in_loop(bsdgo1_outlet.get_details)
#         response = CORRECT_BSDGO1_DETAILS[0]
#         result = response.get('result', {})

#         expected_status = 'on' if result.get('powerSwitch_1') == 1 else 'off'
#         assert bsdgo1_outlet.state.device_status == expected_status

#         assert result.get('active_time') == Defaults.active_time
#         assert result.get('connectionStatus') == 'online'

#     def test_bsdgo1_details_fail(self):
#         """Test BSDGO1 get_details with bad response."""
#         resp_dict, status = BAD_BSDGO1_LIST
#         self.mock_api.return_value = orjson.dumps(resp_dict), status
#         bsdgo1_outlet = VeSyncOutletBSDGO1(
#             self.device_list_model(),
#             self.manager,
#             self.device_map(),
#         )
#         self.run_in_loop(bsdgo1_outlet.get_details)
#         assert len(self.caplog.records) == 1
#         assert 'details' in self.caplog.text

#     def test_bsdgo1_onoff(self):
#         """Test BSDGO1 Device On/Off Methods."""
#         self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
#         bsdgo1_outlet = VeSyncOutletBSDGO1(
#             self.device_list_model(),
#             self.manager,
#             self.device_map(),
#         )
#         head = Helpers.req_header_bypass()
#         body = Helpers.req_body(self.manager, 'bypassV2')
#         body['cid'] = bsdgo1_outlet.cid
#         body['configModule'] = bsdgo1_outlet.config_module

#         # Test turn_on
#         body['payload'] = {
#             'data': {'powerSwitch_1': 1},
#             'method': 'setProperty',
#             'source': 'APP'
#         }
#         on = self.run_in_loop(bsdgo1_outlet.turn_on)
#         self.mock_api.assert_called_with(
#             '/cloud/v2/deviceManaged/bypassV2',
#             'post',
#             headers=head,
#             json_object=body
#         )
#         assert on

#         # Test turn_off
#         body['payload'] = {
#             'data': {'powerSwitch_1': 0},
#             'method': 'setProperty',
#             'source': 'APP'
#         }
#         off = self.run_in_loop(bsdgo1_outlet.turn_off)
#         self.mock_api.assert_called_with(
#             '/cloud/v2/deviceManaged/bypassV2',
#             'post',
#             headers=head,
#             json_object=body
#         )
#         assert off

#     def test_bsdgo1_onoff_fail(self):
#         """Test BSDGO1 On/Off Fail with bad response."""
#         resp_dict, status = BAD_BSDGO1_LIST
#         self.mock_api.return_value = orjson.dumps(resp_dict), status
#         bsdgo1_outlet = VeSyncOutletBSDGO1(
#             self.device_list_model(),
#             self.manager,
#             self.device_map(),
#         )
#         assert not self.run_in_loop(bsdgo1_outlet.turn_on)
#         assert not self.run_in_loop(bsdgo1_outlet.turn_off)



"""Base classes for all VeSync bulbs."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseToggleDevice
from pyvesync.const import BulbFeatures
from pyvesync.utils.colors import HSV, RGB, Color
from pyvesync.utils.helpers import Converters, Validators

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import BulbMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


logger = logging.getLogger(__name__)


class BulbState(DeviceState):
    """VeSync Bulb State Base.

    Base class to hold all state attributes for bulb devices. Inherits from
    [`DeviceState`][pyvesync.base_devices.vesyncbasedevice.DeviceState]. This
    class should not be used directly for devices, but rather subclassed for each
    bulb type.

    Args:
        device (VeSyncBulb): VeSync Bulb device.
        details (ResponseDeviceDetailsModel): Device details from API.
        feature_map (BulbMap): Feature map for bulb.

    Attributes:
        _exclude_serialization (list[str]): List of attributes to exclude from
            serialization.
        active_time (int): Active time of device, defaults to None.
        connection_status (str): Connection status of device.
        device (VeSyncBaseDevice): Device object.
        device_status (str): Device status.
        features (dict): Features of device.
        last_update_ts (int): Last update timestamp of device, defaults to None.
        brightness (int): Brightness of bulb (0-100).
        color_temp (int): White color temperature of bulb in percent (0-100).
        color_temp_kelvin (int): White color temperature of bulb in Kelvin.
        color (Color): Color of bulb in the form of a dataclass with two namedtuple
            attributes - `hsv` & `rgb`. See
            [utils.colors.Colors][pyvesync.utils.colors.Color].
        color_mode (str): Color mode of bulb.
        color_modes (list[str]): List of color modes supported by bulb.

    Methods:
        update_ts: Update last update timestamp.
        to_dict: Dump state to JSON.
        to_json: Dump state to JSON string.
        to_jsonb: Dump state to JSON bytes.
        as_tuple: Convert state to tuple of (name, value) tuples.

    See Also:
        - [`VeSyncBulb`][pyvesync.base_devices.bulb_base.VeSyncBulb]
        - [`ResponseDeviceDetailsModel`][
            pyvesync.models.device_list_models.ResponseDeviceDetailsModel]
        - [`BulbMap`][pyvesync.device_map.BulbMap]
    """

    __slots__ = ('_brightness', '_color', '_color_temp', 'color_mode', 'color_modes')

    def __init__(
        self,
        device: VeSyncBulb,
        details: ResponseDeviceDetailsModel,
        feature_map: BulbMap,
    ) -> None:
        """Initialize VeSync Bulb State Base."""
        super().__init__(device, details, feature_map)
        self._exclude_serialization: list[str] = ['rgb', 'hsv']
        self.features: list[str] = feature_map.features
        self.color_modes: list[str] = feature_map.color_modes
        self.device: VeSyncBulb = device
        self.color_mode: str | None = None
        self._brightness: int | None = None
        self._color_temp: int | None = None
        self._color: Color | None = None

    @property
    def color(self) -> Color | None:
        """Return color of bulb."""
        return self._color

    @color.setter
    def color(self, color: Color | None) -> None:
        """Set color of bulb."""
        self._color = color

    @property
    def hsv(self) -> HSV | None:
        """Return color of bulb as HSV."""
        if self._color is not None:
            return self._color.hsv
        return None

    @hsv.setter
    def hsv(self, hsv_object: HSV | None) -> None:
        """Set color property with HSV values."""
        if hsv_object is None:
            self._color = None
            return
        self._color = Color(hsv_object)

    @property
    def rgb(self) -> RGB | None:
        """Return color of bulb as RGB."""
        if self._color is not None:
            return self._color.rgb
        return None

    @rgb.setter
    def rgb(self, rgb_object: RGB | None) -> None:
        """Set color property with RGB values."""
        if rgb_object is None:
            self._color = None
            return
        self._color = Color(rgb_object)

    @property
    def brightness(self) -> int | None:
        """Brightness of vesync bulb 0-100."""
        if not self.device.supports_brightness:
            logger.warning('Brightness not supported by this bulb')
            return None
        return self._brightness

    @brightness.setter
    def brightness(self, value: int | None) -> None:
        if not self.device.supports_brightness:
            logger.warning('Brightness not supported by this bulb')
            return
        if Validators.validate_zero_to_hundred(value):
            self._brightness = value
        else:
            self._brightness = None

    @property
    def color_temp(self) -> int | None:
        """White color temperature of bulb in percent (0-100)."""
        if not self.device.supports_color_temp:
            logger.warning('Color temperature not supported by this bulb')
            return None
        return self._color_temp

    @color_temp.setter
    def color_temp(self, value: int) -> None:
        if not self.device.supports_color_temp:
            logger.warning('Color temperature not supported by this bulb')
            return
        if Validators.validate_zero_to_hundred(value):
            self._color_temp = value
        else:
            logger.warning('Color temperature must be between 0 and 100')

    @property
    def color_temp_kelvin(self) -> int | None:
        """Return white color temperature of bulb in Kelvin."""
        if not self.device.supports_color_temp:
            logger.warning('Color temperature not supported by this bulb')
            return None
        if self._color_temp is not None:
            return Converters.color_temp_pct_to_kelvin(self._color_temp)
        return None


class VeSyncBulb(VeSyncBaseToggleDevice[BulbState]):
    """Base class for VeSync Bulbs.

    Abstract base class to provide methods for controlling and
    getting details of VeSync bulbs. Inherits from
    [`VeSyncBaseDevice`][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseDevice]. This
    class should not be used directly for devices, but rather subclassed for each bulb.

    Args:
        details (ResponseDeviceDetailsModel): Device details from API.
        manager (VeSync): VeSync API manager.
        feature_map (BulbMap): Feature map for bulb.

    Attributes:
        state (BulbState): Device state object
            Each device has a separate state base class in the base_devices module.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: BulbMap
    ) -> None:
        """Initialize VeSync smart bulb base class."""
        super().__init__(details, manager, feature_map)
        self.state: BulbState = BulbState(self, details, feature_map)

    @property
    def supports_brightness(self) -> bool:
        """Return True if bulb supports brightness."""
        return BulbFeatures.DIMMABLE in self.features

    @property
    def supports_color_temp(self) -> bool:
        """Return True if bulb supports color temperature."""
        return BulbFeatures.COLOR_TEMP in self.features

    @property
    def supports_multicolor(self) -> bool:
        """Return True if bulb supports backlight."""
        return BulbFeatures.MULTICOLOR in self.features

    async def set_hsv(self, hue: float, saturation: float, value: float) -> bool:
        """Set HSV if supported by bulb.

        Args:
            hue (float): Hue 0-360
            saturation (float): Saturation 0-100
            value (float): Value 0-100

        Returns:
            bool: True if successful, False otherwise.
        """
        del hue, saturation, value
        if not self.supports_multicolor:
            logger.debug('Color mode is not supported on this bulb.')
        else:
            logger.debug('set_hsv not configured for %s bulb', self.device_type)
        return False

    async def set_rgb(self, red: float, green: float, blue: float) -> bool:
        """Set RGB if supported by bulb.

        Args:
            red (float): Red 0-255
            green (float): green 0-255
            blue (float): blue 0-255

        Returns:
            bool: True if successful, False otherwise.
        """
        del red, green, blue
        if not self.supports_multicolor:
            logger.debug('Color mode is not supported on this bulb.')
        else:
            logger.debug('set_rgb not configured for %s bulb', self.device_type)
        return False

    async def set_brightness(self, brightness: int) -> bool:
        """Set brightness if supported by bulb.

        Args:
            brightness (NUMERIC_T): Brightness 0-100

        Returns:
            bool: True if successful, False otherwise.
        """
        del brightness
        logger.warning('Brightness not supported/configured by this bulb')
        return False

    async def set_color_temp(self, color_temp: int) -> bool:
        """Set color temperature if supported by bulb.

        Args:
            color_temp (int): Color temperature 0-100

        Returns:
            bool: True if successful, False otherwise.
        """
        del color_temp
        if self.supports_color_temp:
            logger.debug('Color temperature is not configured on this bulb.')
        else:
            logger.debug('Color temperature not supported by this bulb')
        return False

    async def set_white_mode(self) -> bool:
        """Set white mode if supported by bulb.

        Returns:
            bool: True if successful, False otherwise.
        """
        if self.supports_multicolor:
            logger.debug('White mode is not configured on this bulb.')
        else:
            logger.warning('White mode not supported by this bulb')
        return False

    async def set_color_mode(self, color_mode: str) -> bool:
        """Set color mode if supported by bulb.

        Args:
            color_mode (str): Color mode to set.

        Returns:
            bool: True if successful, False otherwise.
        """
        del color_mode
        if self.supports_multicolor:
            logger.debug('Color mode is not configured on this bulb.')
        else:
            logger.warning('Color mode not supported by this bulb')
        return False

    @deprecated('Use `set_white_mode` instead.')
    async def enable_white_mode(self) -> bool:
        """Enable white mode if supported by bulb."""
        return await self.set_white_mode()



"""
Air Purifier and Humidifier Device API Responses

FANS variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'DEV_TYPE': defaultdict(
        lambda: ({"code": 0, "msg": "success"}, 200))
    )
}

# For a function to handle the response
def status_response(request_body=None):
    # do work with request_body
    return request_body, 200

METHOD_RESPONSES['DEV_TYPE']['set_status'] = status_response

# To change the default value for a device type

METHOD_RESPONSES['DEVTYPE'].default_factory = lambda: ({"code": 0, "msg": "success"}, 200)
"""

from copy import deepcopy
from pyvesync.device_map import fan_modules
from pyvesync.const import DeviceStatus
from defaults import (
    build_bypass_v2_response,
    FunctionResponsesV2,
)
from pyvesync.const import FanModes

FANS = [m.setup_entry for m in fan_modules]
FANS_NUM = len(FANS)


class FanDefaults:
    fan_status = DeviceStatus.ON
    screen_status = DeviceStatus.ON
    oscillation_status = DeviceStatus.ON
    oscillation_horizontal = DeviceStatus.ON
    oscillation_vertical = DeviceStatus.ON
    mute_status = DeviceStatus.ON
    child_lock = DeviceStatus.OFF
    fan_mode = FanModes.NORMAL
    yaw = 45
    pitch = 90
    top = 120
    bottom = 0
    left = 0
    right = 90
    fan_level = 1
    fan_speed_level = 1
    temperature_fan = 75.0


FAN_DETAILS: dict[str, dict] = {
    "LTF-F422S": {
        "powerSwitch": int(FanDefaults.fan_status),
        "workMode": FanModes.NORMAL.value,
        "manualSpeedLevel": FanDefaults.fan_level,
        "fanSpeedLevel": FanDefaults.fan_speed_level,
        "screenState": int(FanDefaults.screen_status),
        "screenSwitch": int(FanDefaults.screen_status),
        "oscillationSwitch": int(FanDefaults.oscillation_status),
        "oscillationState": int(FanDefaults.oscillation_status),
        "muteSwitch": int(FanDefaults.mute_status),
        "muteState": int(FanDefaults.mute_status),
        "timerRemain": 0,
        "temperature": FanDefaults.temperature_fan,
        "sleepPreference": {
            "sleepPreferenceType": "default",
            "oscillationSwitch": int(FanDefaults.oscillation_status),
            "initFanSpeedLevel": 0,
            "fallAsleepRemain": 0,
            "autoChangeFanLevelSwitch": int(FanDefaults.fan_status),
        },
        "scheduleCount": 0,
        "displayingType": 0,
        "errorCode": 0,
    },
    "LPF-R423S": {
        "powerSwitch": int(FanDefaults.fan_status),
        "workMode": FanDefaults.fan_mode.value,
        "fanSpeedLevel": FanDefaults.fan_speed_level,
        "scheduleCount": 0,
        "timerRemain": 0,
        "sleepPreference": {
            "sleepPreferenceType": "default",
            "oscillationState": int(FanDefaults.oscillation_status),
            "initFanSpeedLevel": 0,
            "fallAsleepRemain": 0,
        },
        "temperature": FanDefaults.temperature_fan * 10,
        "muteState": int(FanDefaults.mute_status),
        "muteSwitch": int(FanDefaults.mute_status),
        "screenState": int(FanDefaults.screen_status),
        "screenSwitch": int(FanDefaults.screen_status),
        "errorCode": 0,
        "horizontalOscillationState": int(FanDefaults.oscillation_horizontal),
        "verticalOscillationState": int(FanDefaults.oscillation_vertical),
        "childLock": FanDefaults.child_lock,
        "oscillationCoordinate": {"yaw": FanDefaults.yaw, "pitch": FanDefaults.pitch},
        "oscillationRange": {
            "left": FanDefaults.left,
            "right": FanDefaults.right,
            "top": FanDefaults.top,
            "bottom": FanDefaults.bottom,
        },
        "highTemperatureReminderState": 1,
        "highTemperature": 806,
        "smartCleaningReminderState": 1,
        "oscillationCalibrationState": 0,
        "oscillationCalibrationProgress": 0,
        "levelMemory": [
            {"workMode": "normal", "level": 2, "enable": 1},
            {"workMode": "turbo", "level": 12, "enable": 0},
            {"workMode": "eco", "level": 3, "enable": 1},
            {"workMode": "advancedSleep", "level": 3, "enable": 1},
        ],
        "horizontalOscillationDemo": 0,
        "verticalOscillationDemo": 0,
        "isSupportSetOnceOscillation": 1,
        "isTimerSupportPowerOn": 1,
        "isSupportSetRelativeCoordinate": 1,
    },
}
"""Contains the result dictionary of the device response details for fans.

Dictionary with key being the `setup_entry` attribute for each fan device
found in the `pyvesync.device_map.fan_modules` object. The value is the
return dictionary for a successful `fan.update()` API request.
"""

DETAILS_RESPONSES = {
    "LTF-F422S": build_bypass_v2_response(inner_result=FAN_DETAILS["LTF-F422S"]),
    "LPF-R423S": build_bypass_v2_response(inner_result=FAN_DETAILS["LPF-R423S"]),
}


METHOD_RESPONSES = {
    "LTF-F422S": deepcopy(FunctionResponsesV2),
    "LPF-R423S": deepcopy(FunctionResponsesV2)
}
"""Default responses for device methods.

This dictionary maps the setup_entry to the default response for that device.

Examples:
    - For legacy API responses `defaults.FunctionResponses`
    - For Bypass V1 `defaults.FunctionResponsesV1`
    - For Bypass V2 `defaults.FunctionResponsesV2`
"""

# Add responses for methods with different responses than the default



"""Helper functions for VeSync API."""

from __future__ import annotations

import hashlib
import logging
import re
import time
from collections.abc import Iterator
from dataclasses import InitVar, dataclass, field
from enum import StrEnum
from functools import lru_cache
from typing import TYPE_CHECKING, Any, TypeVar

from mashumaro.exceptions import InvalidFieldValue, MissingField, UnserializableField
from mashumaro.mixins.orjson import DataClassORJSONMixin
from typing_extensions import deprecated

from pyvesync.const import (
    APP_VERSION,
    BYPASS_HEADER_UA,
    DEFAULT_REGION,
    KELVIN_MAX,
    KELVIN_MIN,
    MOBILE_ID,
    PHONE_BRAND,
    PHONE_OS,
    USER_TYPE,
    ConnectionStatus,
)
from pyvesync.models.base_models import DefaultValues
from pyvesync.utils.errors import ErrorCodes, ErrorTypes, ResponseInfo
from pyvesync.utils.logs import LibraryLogger

if TYPE_CHECKING:
    from pyvesync.base_devices.vesyncbasedevice import VeSyncBaseDevice
    from pyvesync.vesync import VeSync


T = TypeVar('T')
T_MODEL = TypeVar('T_MODEL', bound=DataClassORJSONMixin)

ST = TypeVar('ST', str, int)

_LOGGER = logging.getLogger(__name__)

NUMERIC_OPT = float | str | None

NUMERIC_STRICT = float | str

REQUEST_T = dict[str, Any]


class Validators:
    """Methods to validate input."""

    @staticmethod
    def validate_range(
        value: NUMERIC_OPT, minimum: NUMERIC_STRICT, maximum: NUMERIC_STRICT
    ) -> bool:
        """Validate number is within range."""
        if value is None:
            return False
        try:
            return float(minimum) <= float(value) <= float(maximum)
        except (ValueError, TypeError):
            return False

    @classmethod
    def validate_zero_to_hundred(cls, value: NUMERIC_OPT) -> bool:
        """Validate number is a percentage."""
        return Validators.validate_range(value, 0, 100)

    @classmethod
    def validate_hsv(
        cls, hue: NUMERIC_OPT, saturation: NUMERIC_OPT, value: NUMERIC_OPT
    ) -> bool:
        """Validate HSV values."""
        return (
            cls.validate_range(hue, 0, 360)
            and cls.validate_zero_to_hundred(saturation)
            and cls.validate_zero_to_hundred(value)
        )

    @classmethod
    def validate_rgb(
        cls, red: NUMERIC_OPT, green: NUMERIC_OPT, blue: NUMERIC_OPT
    ) -> bool:
        """Validate RGB values."""
        return all(cls.validate_range(val, 0, 255) for val in (red, green, blue))


class Converters:
    """Helper functions to convert units."""

    @staticmethod
    def color_temp_kelvin_to_pct(kelvin: int) -> int:
        """Convert Kelvin to percentage."""
        return int((kelvin - KELVIN_MIN) / (KELVIN_MAX - KELVIN_MIN) * 100)

    @staticmethod
    def color_temp_pct_to_kelvin(pct: int) -> int:
        """Convert percentage to Kelvin."""
        return int(KELVIN_MIN + ((pct / 100) * (KELVIN_MAX - KELVIN_MIN)))

    @staticmethod
    def temperature_kelvin_to_celsius(kelvin: int) -> float:
        """Convert Kelvin to Celsius."""
        return kelvin - 273.15

    @staticmethod
    def temperature_celsius_to_kelvin(celsius: float) -> int:
        """Convert Celsius to Kelvin."""
        return int(celsius + 273.15)

    @staticmethod
    def temperature_fahrenheit_to_celsius(fahrenheit: float) -> float:
        """Convert Fahrenheit to Celsius."""
        return (fahrenheit - 32) * 5.0 / 9.0

    @staticmethod
    def temperature_celsius_to_fahrenheit(celsius: float) -> float:
        """Convert Celsius to Fahrenheit."""
        return celsius * 9.0 / 5.0 + 32


class Helpers:
    """VeSync Helper Functions."""

    @staticmethod
    def model_maker(
        logger: logging.Logger,
        model: type[T_MODEL],
        method_name: str,
        data: dict[str, Any],
        device: VeSyncBaseDevice | None = None,
    ) -> T_MODEL | None:
        """Create a model instance from a dictionary.

        This method catches common errors that occur when parsing the
        API response and returns None if the data is invalid. Enable debug
        or verbose logging to see more information.


        Args:
            logger (logging.Logger): Logger instance.
            model (type[T_MODEL]): Model class to create an instance of.
            method_name (str): Name of the method used in API call.
            device (VeSyncBaseDevice): Instance of VeSyncBaseDevice.
            data (dict[str, Any] | None): Dictionary to create the model from.

        Returns:
            T_MODEL: Instance of the model class.
        """
        try:
            model_instance = model.from_dict(data)
        except (MissingField, UnserializableField, InvalidFieldValue) as err:
            LibraryLogger.error_mashumaro_response(
                logger,
                method_name,
                data,
                err,
                device,
            )
            return None
        return model_instance

    @staticmethod
    def bump_level(level: T | None, levels: list[T]) -> T:
        """Increment level by one returning to first level if at last.

        Args:
            level (T | None): Current level.
            levels (list[T]): List of levels.
        """
        if level in levels:
            idx = levels.index(level)
            if idx < len(levels) - 1:
                return levels[idx + 1]
        return levels[0]

    @classmethod
    def process_dev_response(
        cls,
        logger: logging.Logger,
        method_name: str,
        device: VeSyncBaseDevice,
        r_dict: dict | None,
    ) -> dict | None:
        """Process JSON response from Bytes.

        Parses bytes and checks for errors common to all JSON
        responses, included checking the "code" key for non-zero
        values. Outputs error to passed logger with formatted string
        if an error is found. This also saves the response code information
        to the `device.last_response` attribute.

        Args:
            logger (logging.Logger): Logger instance.
            method_name (str): Method used in API call.
            r_dict (dict | None): JSON response from API.
            device (VeSyncBaseDevice): Instance of VeSyncBaseDevice.

        Returns:
            dict | None: Parsed JSON response or None if there was an error.
        """
        device.state.update_ts()
        if r_dict is None:
            LibraryLogger.error_device_response_content(
                logger,
                device,
                method_name,
                'No response or cannot deserialize response from API',
            )
            device.last_response = ResponseInfo(
                name='INVALID_RESPONSE',
                error_type=ErrorTypes.BAD_RESPONSE,
                message=f'No response from API for {method_name}',
            )
            return None

        error_info = cls.parse_error_code(r_dict)
        if error_info.device_online is False:
            device.state.connection_status = ConnectionStatus.OFFLINE
        else:
            device.state.connection_status = ConnectionStatus.ONLINE
        LibraryLogger.log_device_return_code(
            logger,
            method_name,
            device.device_name,
            device.device_type,
            error_info,
        )
        device.last_response = error_info
        device.last_response.response_data = r_dict
        if error_info.code != 0:
            return None
        return r_dict

    @staticmethod
    def get_class_attributes(target_class: object, keys: tuple[str]) -> dict[str, Any]:
        """Find matching attributes, static methods, and class methods from list of keys.

        This function is case insensitive and will remove underscores from the keys before
        comparing them to the class attributes. The provided keys will be returned in the
        same format if found

        Args:
            target_class (object): Class to search for attributes
            keys (tuple[str]): Tuple of keys to search for

        Returns:
            dict[str, Any]: Dictionary of keys and their values from the class
        """
        alias_map = {
            'userCountryCode': 'countrycode',
            'deviceId': 'cid',
            'homeTimeZone': 'timezone',
            'configModel': 'configmodule',
            'region': 'countrycode',
        }

        def normalize_name(name: str) -> str:
            """Normalize a string by removing underscores and making it lowercase."""
            return re.sub(r'_', '', name).lower()

        def get_value(attr_name: str) -> str | float | None:
            """Get value from attribute."""
            attr = getattr(target_class, attr_name)
            try:
                return attr() if callable(attr) else attr  # type: ignore[no-any-return]
            except TypeError:
                return None

        result = {}
        normalized_keys = {normalize_name(key): key for key in keys}
        normalized_aliases = [normalize_name(key) for key in alias_map.values()]

        for attr_name in dir(target_class):
            normalized_attr = normalize_name(attr_name)
            if normalized_attr in normalized_keys:
                attr_val = get_value(attr_name)
                if attr_val is not None:
                    result[normalized_keys[normalized_attr]] = attr_val
            if normalized_attr in normalized_aliases:
                attr_val = get_value(attr_name)
                if attr_val is not None:
                    key_index = normalized_aliases.index(normalized_attr)
                    key_val = list(alias_map.keys())[key_index]
                    if key_val in keys:
                        result[key_val] = attr_val

        return result

    @classmethod
    @lru_cache(maxsize=128, typed=True)
    def get_defaultvalues_attributes(cls, keys: tuple[str]) -> dict[str, Any]:
        """Get default values for dataclass attributes.

        This method uses lru_cache to cache results for faster subsequent lookups.

        Args:
            keys (list[str]): List of attribute names to get default values for.

        Returns:
            dict[str, Any]: Dictionary of attribute names and their default values.
        """
        return cls.get_class_attributes(DefaultValues, keys)

    @classmethod
    @lru_cache(maxsize=128, typed=True)
    def get_manager_attributes(cls, manager: VeSync, keys: tuple[str]) -> dict[str, Any]:
        """Get VeSync manager attributes.

        This method uses lru_cache to cache results for faster subsequent lookups.

        Args:
            manager (VeSyncManager): Instance of VeSyncManager.
            keys (list[str]): List of attribute names to get values for.

        Returns:
            dict[str, Any]: Dictionary of attribute names and their values.
        """
        return cls.get_class_attributes(manager, keys)

    @classmethod
    @lru_cache(maxsize=128, typed=True)
    def get_device_attributes(
        cls, device: VeSyncBaseDevice, keys: tuple[str]
    ) -> dict[str, Any]:
        """Get VeSync device attributes.

        This method uses lru_cache to cache results for faster subsequent lookups.

        Args:
            device (VeSyncBaseDevice): Instance of VeSyncBaseDevice.
            keys (list[str]): List of attribute names to get values for.

        Returns:
            dict[str, Any]: Dictionary of attribute names and their values.
        """
        return cls.get_class_attributes(device, keys)

    @staticmethod
    def req_legacy_headers(manager: VeSync) -> dict[str, str]:
        """Build header for legacy api GET requests.

        Args:
            manager (VeSyncManager): Instance of VeSyncManager.

        Returns:
            dict: Header dictionary for api requests.

        Examples:
            >>> req_headers(manager)
            {
                'accept-language': 'en',
                'accountId': manager.account_id,
                'appVersion': APP_VERSION,
                'content-type': 'application/json',
                'tk': manager.token,
                'tz': manager.time_zone,
            }

        """
        return {
            'accept-language': 'en',
            'accountId': manager.account_id,  # type: ignore[dict-item]
            'appVersion': APP_VERSION,
            'content-type': 'application/json',
            'tk': manager.token,  # type: ignore[dict-item]
            'tz': manager.time_zone,
        }

    @staticmethod
    def req_header_bypass() -> dict[str, str]:
        """Build header for api requests on 'bypass' endpoint.

        Returns:
            dict: Header dictionary for api requests.

        Examples:
            >>> req_header_bypass()
            {
                'Content-Type': 'application/json; charset=UTF-8',
                'User-Agent': BYPASS_HEADER_UA,
            }
        """
        return {
            'Content-Type': 'application/json; charset=UTF-8',
            'User-Agent': BYPASS_HEADER_UA,
        }

    @staticmethod
    def _req_body_base(manager: VeSync) -> dict[str, str]:
        """Return universal keys for body of api requests.

        Args:
            manager (VeSyncManager): Instance of VeSyncManager.

        Returns:
            dict: Body dictionary for api requests.

        Examples:
            >>> req_body_base(manager)
            {
                'timeZone': manager.time_zone,
                'acceptLanguage': 'en',
            }
        """
        return {'timeZone': manager.time_zone, 'acceptLanguage': 'en'}

    @staticmethod
    def _req_body_auth(manager: VeSync) -> REQUEST_T:
        """Keys for authenticating api requests.

        Args:
            manager (VeSyncManager): Instance of VeSyncManager.

        Returns:
            dict: Authentication keys for api requests.

        Examples:
            >>> req_body_auth(manager)
            {
                'accountID': manager.account_id,
                'token': manager.token,
            }
        """
        return {'accountID': manager.account_id, 'token': manager.token}

    @staticmethod
    @deprecated('This is a legacy function and will be removed in a future release.')
    def _req_body_details() -> REQUEST_T:
        """Detail keys for api requests.

        This method is deprecated, use `get_class_attributes` instead.

        Returns:
            dict: Detail keys for api requests.

        Examples:
            >>> req_body_details()
            {
                'appVersion': APP_VERSION,
                'phoneBrand': PHONE_BRAND,
                'phoneOS': PHONE_OS,
                'traceId': str(int(time.time())),
            }
        """
        return {
            'appVersion': APP_VERSION,
            'phoneBrand': PHONE_BRAND,
            'phoneOS': PHONE_OS,
            'traceId': str(int(time.time())),
        }

    @classmethod
    @deprecated('This is a legacy function and will be removed in a future release.')
    def req_body(cls, manager: VeSync, type_: str) -> REQUEST_T:  # noqa: C901
        """Builder for body of api requests.

        This method is deprecated, use `get_class_attributes` instead.

        Args:
            manager (VeSyncManager): Instance of VeSyncManager.
            type_ (str): Type of request to build body for.

        Returns:
            dict: Body dictionary for api requests.

        Note:
            The body dictionary will be built based on the type of request.
            The type of requests include:
            - login
            - devicestatus
            - devicelist
            - devicedetail
            - energy_week
            - energy_month
            - energy_year
            - bypass
            - bypassV2
            - bypass_config
        """
        body: REQUEST_T = cls._req_body_base(manager)

        if type_ == 'login':
            body |= cls._req_body_details()
            # pylint: disable=protected-access
            body |= {
                'email': manager.auth._username,  # noqa: SLF001
                'password': cls.hash_password(
                    manager.auth._password  # noqa: SLF001
                ),
                'devToken': '',
                'userType': USER_TYPE,
                'method': 'login',
            }
            return body

        body |= cls._req_body_auth(manager)

        if type_ == 'devicestatus':
            return body

        body |= cls._req_body_details()

        if type_ == 'devicelist':
            body['method'] = 'devices'
            body['pageNo'] = '1'
            body['pageSize'] = '100'

        elif type_ == 'devicedetail':
            body['method'] = 'devicedetail'
            body['mobileId'] = MOBILE_ID

        elif type_ == 'energy_week':
            body['method'] = 'energyweek'
            body['mobileId'] = MOBILE_ID

        elif type_ == 'energy_month':
            body['method'] = 'energymonth'
            body['mobileId'] = MOBILE_ID

        elif type_ == 'energy_year':
            body['method'] = 'energyyear'
            body['mobileId'] = MOBILE_ID

        elif type_ == 'bypass':
            body['method'] = 'bypass'

        elif type_ == 'bypassV2':
            body['deviceRegion'] = DEFAULT_REGION
            body['method'] = 'bypassV2'

        elif type_ == 'bypass_config':
            body['method'] = 'firmwareUpdateInfo'

        return body

    @staticmethod
    def calculate_hex(hex_string: str) -> float:
        """Credit for conversion to itsnotlupus/vesync_wsproxy.

        Hex conversion for legacy outlet power and voltage.
        """
        hex_conv = hex_string.split(':')
        return (int(hex_conv[0], 16) + int(hex_conv[1], 16)) / 8192

    @staticmethod
    def hash_password(string: str) -> str:
        """Encode password."""
        return hashlib.md5(string.encode('utf-8')).hexdigest()  # noqa: S324

    @staticmethod
    def parse_error_code(response: dict) -> ResponseInfo:  # noqa: C901
        """Get all error codes from nested dictionary.

        Args:
            response (dict): API response.

        Returns:
            list[int]: List of error codes.
        """
        error_keys = ['error', 'code', 'error_code', 'errorCode']

        def extract_all_error_codes(
            key: str, var: dict
        ) -> Iterator[tuple[int, str | None]]:
            """Find all error code keys in nested dictionary."""
            if hasattr(var, 'items'):
                for k, v in var.items():
                    if k == key and isinstance(v, int) and v != 0:
                        msg = var.get('msg') or var.get('result', {}).get('msg')
                        yield v, msg
                    if isinstance(v, dict):
                        yield from extract_all_error_codes(key, v)
                    elif isinstance(v, list):
                        for item in v:
                            yield from extract_all_error_codes(key, item)

        errors: list[tuple[int, str | None]] = []

        if 'code' in response:
            outside_code = response.get('code', '')
            outside_msg = response.get('msg')

            if str(outside_code) != '0':
                return ErrorCodes.get_error_info(int(outside_code), outside_msg)

        for error_key in error_keys:
            errors.extend(list(extract_all_error_codes(error_key, response)))
        return_code = 0
        return_msg = None
        for code, msg in errors:
            if code != 0:
                return_code = code
                return_msg = msg
        return ErrorCodes.get_error_info(int(return_code), return_msg)

    @staticmethod
    def get_key(
        data: dict[StrEnum | str, ST], value: ST, default: str | None = None
    ) -> str | StrEnum | None:
        """Get key from dictionary ignoring case sensitivity.

        Args:
            data (dict[str, Any]): Dictionary to search.
            value (str): Value to search for.
            default (Any): Default value to return if key not found.

        Returns:
            Any: Value associated with the key, or None if not found.
        """
        return next((k for k, v in data.items() if v == value), default)


@dataclass(repr=False)
class Timer:
    """Dataclass to hold state of timers.

    Note:
        This should be used by VeSync device instances to manage internal status,
        does not interact with the VeSync API.

    Args:
        timer_duration (int): Length of timer in seconds
        action (str): Action to perform when timer is done
        id (int): ID of timer, defaults to 1
        remaining (int): Time remaining on timer in seconds, defaults to None
        _update_time (int): Last updated unix timestamp in seconds, defaults to None

    Attributes:
        update_time (str): Timestamp of last update
        status (str): Status of timer, one of 'active', 'paused', 'done'
        time_remaining (int): Time remaining on timer in seconds
        running (bool): True if timer is running
        paused (bool): True if timer is paused
        done (bool): True if timer is done
    """

    timer_duration: int
    action: str
    id: int = 1
    remaining: InitVar[int | None] = None
    _status: str = field(default='active', init=False, repr=False)
    _remain: int = field(default=0, init=False, repr=False)
    _update_time: int = int(time.time())

    def __post_init__(self, remaining: int | None) -> None:
        """Set remaining time if provided."""
        if remaining is not None:
            self._remain = remaining
        else:
            self._remain = self.timer_duration

    def __repr__(self) -> str:
        """Return string representation of the Timer object.

        Returns:
            str: String representation of the Timer object.
        """
        return (
            f'Timer(id={self.id}, duration={self.timer_duration}, '
            f'status={self.status}, remaining={self.time_remaining})'
        )

    def update_ts(self) -> None:
        """Update timestamp."""
        self._update_time = int(time.time())

    @property
    def status(self) -> str:
        """Return status of timer."""
        if self._status in ('paused', 'done'):
            return self._status
        if self.time_remaining <= 0:
            self._status = 'done'
            return 'done'
        return 'active'

    @property
    def time_remaining(self) -> int:
        """Return remaining seconds."""
        if self._status == 'paused':
            return self._remain
        if self._status == 'done':
            return 0

        # 'active' state - compute how much time has ticked away
        elapsed = time.time() - self._update_time
        current_remaining = self._remain - elapsed

        # If we've run out of time, mark it done
        if current_remaining <= 0:
            return 0
        return int(current_remaining)

    @property
    def running(self) -> bool:
        """Check if timer is active."""
        return self.time_remaining > 0 and self.status == 'active'

    @property
    def paused(self) -> bool:
        """Check if timer is paused."""
        return bool(self.status == 'paused')

    @property
    def done(self) -> bool:
        """Check if timer is complete."""
        return bool(self.time_remaining <= 0 or self._status == 'done')

    def end(self) -> None:
        """Change status of timer to done."""
        self._status = 'done'
        self._remain = 0

    def start(self) -> None:
        """Restart paused timer."""
        if self._status != 'paused':
            return
        self._update_time = int(time.time())
        self._status = 'active'

    def pause(self) -> None:
        """Pauses the timer if it's active.

        Performs the following steps:
            - Calculate the up-to-date remaining time,
            - Update internal counters,
            - Set _status to 'paused'.
        """
        if self._status == 'active':
            # Update the time_remaining based on elapsed
            current_remaining = self.time_remaining
            if current_remaining <= 0:
                self._status = 'done'
                self._remain = 0
            else:
                self._status = 'paused'
                self._remain = current_remaining
            self._update_time = int(time.time())


@dataclass(kw_only=True)
class OscillationCoordinates:
    """Dataclass to hold oscillation coordinates.

    Note:
        This should be used by VeSync device instances to manage internal status,
        does not interact with the VeSync API.
    """

    yaw: int
    pitch: int


@dataclass(kw_only=True)
class OscillationRange:
    """Dataclass to hold oscillation range.

    Note:
        This should be used by VeSync device instances to manage internal status,
        does not interact with the VeSync API.
    """

    left: int
    right: int
    top: int
    bottom: int



"""Common base class for tests and Default values.

Routine Listings
----------------

FunctionResponses : defaultdict
    Defaultdict of the standard response tuple for device methods
Defaults: class
    Default values and methods for generating default values
"""
from typing import TypeVar, Any
from collections import defaultdict
from requests.structures import CaseInsensitiveDict
import pyvesync.const as const
from pyvesync.utils.helpers import Converters
from pyvesync.device_map import DeviceMapTemplate
from pyvesync.utils.colors import Color, RGB

T = TypeVar("T", bound=DeviceMapTemplate)

FunctionResponses: defaultdict = defaultdict(lambda: {"code": 0, "msg": None})

FunctionResponsesV1: defaultdict = defaultdict(
    lambda: {"traceId": TestDefaults.trace_id, "code": 0, "msg": None, "result": {}}
)

FunctionResponsesV2: defaultdict = defaultdict(
    lambda: {
        "traceId": TestDefaults.trace_id,
        "code": 0,
        "msg": None,
        "result": {"code": 0, "msg": None, "traceId": TestDefaults.trace_id, "result": {}},
    }
)

CALL_API_ARGS = ['url', 'method', 'json_object', 'headers']

ID_KEYS = ['CID', 'UUID', 'MACID']


def clean_name(name: str) -> str:
    """Clean the device name by removing unwanted characters."""
    return name.strip().lower().replace(" ", "_").replace("-", "_").replace(".", "_")


def build_base_response(
    code: int = 0, msg: str | None = None, merge_dict: dict | None = None
) -> dict:
    """Build the standard response response tuple."""
    resp_dict = {
        "code": code,
        "msg": msg,
        "stacktrace": None,
        "module": None,
        "traceId": TestDefaults.trace_id,
    }
    resp_dict |= merge_dict or {}
    return resp_dict


def build_bypass_v1_response(
    *,
    code: int = 0,
    msg: str | None = None,
    result_dict: dict | None = None,
    result_override: dict | None = None,
) -> dict:
    """Build the standard response response tuple.

    This function has kw-only arguments.

    Args:
        code (int): The response code.
        msg (str | None): The response message.
        result_dict (dict | None): The result dictionary.
        result_override (dict | None): dictionary to override the supplied result_dict.

    Returns:
        tuple[dict, int]: The response dictionary and status code.
    """
    resp_dict = {
        "code": code,
        "msg": msg,
        "stacktrace": None,
        "module": None,
        "traceId": TestDefaults.trace_id,
        "result": result_dict or {},
    }
    return resp_dict


def build_bypass_v2_response(
    *,
    code: int = 0,
    msg: str | None = None,
    inner_code: int = 0,
    inner_result: dict[str, Any] | None = None,
    inner_result_override: dict | None = None
) -> dict:
    """Build the standard response response tuple for BypassV2 endpoints.

    This function has kw-only arguments.

    Args:
        code(int): The response code.
        msg(str | None): The response message.
        inner_code(int): The inner response code.
        inner_result(dict | None): The inner response result.
        inner_result_override(dict | None): dictionary to override the supplied inner_result.

    Returns:
        tuple[dict, int]: The response dictionary and status code.
    """
    inner_result = inner_result or {}
    resp_dict = build_base_response(code, msg)
    if inner_result_override is not None:
        inner_result = inner_result | inner_result_override
    resp_dict['result'] = {
        "traceId": TestDefaults.trace_id,
        "code": inner_code,
        "result": inner_result,
    }
    return resp_dict


def get_device_map_from_setup_entry(module_list: list[T], setup_entry: str) -> T | None:
    """Get the device map from the setup entry.

    Args:
        module_list (list[T]): The list of device modules.
        setup_entry (str): The setup entry string.

    Returns:
        dict: The device map dictionary.
    """
    for module in module_list:
        if clean_name(module.setup_entry) == clean_name(setup_entry):
            return module
    return None


class TestDefaults:
    """General defaults for API responses and requests.

    Attributes
    ----------
    token : str
        Default token for API requests
    active_time : str
        Default active time for API responses
    account_id : str
        Default account ID for API responses
    active_time : str
        Default active time for API responses
    color: Color (dataclass)
        Red=50, Green=100, Blue=225, Hue=223, Saturation=77.78, value=88.24
        Default Color dataclass
        contains red, green, blue, hue, saturation and value attributes

    Methods
    --------
    name(dev_type='NA')
        Default device name created from "dev_type-NAME"
    cid(dev_type='NA')
        Default device cid created from "dev_type-CID"
    uuid(dev_type='NA')
        Default device uuid created from "dev_type-UUID"
    macid(dev_type='NA')
        Default device macid created from "dev_type-MACID"
    """

    token = 'sample_tk'
    account_id = 'sample_id'
    trace_id = "TRACE_ID"
    terminal_id = "TERMINAL_ID"
    app_id = "APP_ID"
    email = 'EMAIL'
    password = 'PASSWORD'
    authorization_code = "AUTHORIZATION_CODE"
    biz_token = "BIZ_TOKEN"
    region = 'US'
    country_code = 'US'
    active_time = 1
    time_zone = const.DEFAULT_TZ
    color: Color = Color(RGB(50, 100, 225))
    brightness = 100
    color_temp = 100
    color_temp_k = Converters.color_temp_pct_to_kelvin(100)
    bool_toggle = True
    str_toggle = 'on'
    bin_toggle = 1
    level_int = 1
    country_code = 'US'

    @staticmethod
    def name(setup_entry: str = 'NA'):
        """Name of device with format f"{setup_entry}-NAME".

        Parameters
        ----------
        setup_entry : str
            Device type use to create default name

        Returns
        -------
        str
            Default name for device f"{setup_entry}-NAME"
        """
        return f'{setup_entry}-NAME'

    @staticmethod
    def cid(setup_entry='NA'):
        """CID for a device with format f"{setup_entry}-CID".

        Parameters
        ----------
        setup_entry : str
            Device type use to create default cid

        Returns
        -------
        str
            Default cid for device f"setup_entry-CID"
        """
        return f'{setup_entry}-CID'

    @staticmethod
    def uuid(setup_entry='NA'):
        """UUID for a device with format f"{setup_entry}-UUID".

        Parameters
        ----------
        setup_entry : str
            Device type use to create default UUID

        Returns
        -------
        str
            Default uuid for device f"{setup_entry}-UUID"
        """
        return f'{setup_entry}-UUID'

    @staticmethod
    def macid(setup_entry='NA'):
        """MACID for a device with format f"{setup_entry}-MACID".

        Parameters
        ----------
        setup_entry : str
            Device type use to create default macid

        Returns
        -------
        str
            Default macID for device f"{setup_entry}-MACID"
        """
        return f'{setup_entry}-MACID'

    @staticmethod
    def config_module(setup_entry='NA'):
        """Config module for a device with format f"{setup_entry}-CONFIG".

        Parameters
        ----------
        setup_entry : str
            Device type use to create default config module

        Returns
        -------
        str
            Default config module for device f"{setup_entry}-CONFIG_MODULE"
        """
        return f'{setup_entry}-CONFIG_MODULE'


API_DEFAULTS = CaseInsensitiveDict({
    'accountID': TestDefaults.account_id,
    'token': TestDefaults.token,
    'timeZone': const.DEFAULT_TZ,
    'acceptLanguage': 'en',
    'appVersion': const.APP_VERSION,
    'phoneBrand': const.PHONE_BRAND,
    'phoneOS': const.PHONE_OS,
    'userType': const.USER_TYPE,
    "tk": TestDefaults.token,
    "mobileId": "MOBILE_ID",
    "traceId": "TRACE_ID",
    'verifyEmail': 'EMAIL',
    'nickName': 'NICKNAME',
    'password': 'PASSWORD',
    'username': 'EMAIL',
    'email': 'EMAIL',
    'deviceName': 'NAME'

})



"""Etekcity/Valceno Smart Light Bulbs.

This module provides classes for the following Etekcity/Valceno smart lights:

    1. ESL100: Dimmable Bulb
    2. ESL100CW: Tunable White Bulb
    3. XYD0001: RGB Bulb
    4. ESL100MC: Multi-Color Bulb

The classes all inherit from VeSyncBulb, which is a subclass of VeSyncBaseDevice and
[VeSyncToggleDevice][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseToggleDevice].
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices import VeSyncBulb
from pyvesync.const import ConnectionStatus, DeviceStatus
from pyvesync.models import bulb_models
from pyvesync.models.bypass_models import TimerModels
from pyvesync.utils.colors import Color
from pyvesync.utils.device_mixins import (
    BypassV1Mixin,
    BypassV2Mixin,
    process_bypassv1_result,
    process_bypassv2_result,
)
from pyvesync.utils.helpers import Helpers, Timer, Validators

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import BulbMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel

logger = logging.getLogger(__name__)

NUMERIC_OPT = int | float | str | None
NUMERIC_STRICT = float | int | str


class VeSyncBulbESL100MC(BypassV2Mixin, VeSyncBulb):
    """Etekcity ESL100 Multi Color Bulb device.

    Inherits from [VeSyncBulb][pyvesync.base_devices.bulb_base.VeSyncBulb]
    and [VeSyncBaseDevice][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseDevice].

    The state of the bulb is stored in the `state` attribute, which is an of
    [BulbState][pyvesync.base_devices.bulb_base.BulbState]. The `state` attribute
    contains all settable states for the bulb.

    Args:
        details (dict): Dictionary of bulb state details.
        manager (VeSync): Manager class used to make API calls.
        feature_map (BulbMap): Device configuration map.

    Attributes:
        state (BulbState): Device state object
            Each device has a separate state base class in the base_devices module.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.

    Notes:
        The details dictionary contains the device information retrieved by the
        `update()` method:
        ```python
        details = {
            'brightness': 50,
            'colorMode': 'rgb',
            'color' : Color(red=0, green=0, blue=0)
        }
        ```
        See pyvesync.helpers.color.Color for more information on the Color dataclass.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: BulbMap
    ) -> None:
        """Instantiate ESL100MC Multicolor Bulb."""
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getLightStatus')

        result_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, bulb_models.ResponseESL100MCResult
        )
        if result_model is None:
            return
        self._set_state(result_model)
        return

    def _set_state(self, response: bulb_models.ResponseESL100MCResult) -> None:
        """Build detail dictionary from response."""
        self.state.brightness = response.brightness
        self.state.color_mode = response.colorMode
        self.state.color = Color.from_rgb(
            red=response.red, green=response.green, blue=response.blue
        )

    async def set_brightness(self, brightness: int) -> bool:
        return await self.set_status(brightness=brightness)

    async def set_rgb(self, red: float, green: float, blue: float) -> bool:
        return await self.set_status(red=red, green=green, blue=blue)

    async def set_hsv(self, hue: float, saturation: float, value: float) -> bool:
        hsv = Color.from_hsv(hue=hue, saturation=saturation, value=value)
        if hsv is not None:
            return await self.set_status(
                red=hsv.rgb.red, green=hsv.rgb.green, blue=hsv.rgb.blue
            )
        logger.warning('Invalid HSV values')
        return False

    async def set_white_mode(self) -> bool:
        return await self.set_status(brightness=100)

    async def set_status(
        self,
        brightness: float | None = None,
        red: float | None = None,
        green: float | None = None,
        blue: float | None = None,
    ) -> bool:
        """Set color of VeSync ESL100MC.

        Brightness or RGB values must be provided. If RGB values are provided,
        brightness is ignored.

        Args:
            brightness (float | None): Brightness of bulb (0-100).
            red (float | None): Red value of RGB color, 0-255.
            green (float | None): Green value of RGB color, 0-255.
            blue (float | None): Blue value of RGB color, 0-255.

        Returns:
            bool: True if successful, False otherwise.
        """
        brightness_update = 100
        if red is not None and green is not None and blue is not None:
            new_color = Color.from_rgb(red, green, blue)
            color_mode = 'color'
            if (
                self.state.device_status == DeviceStatus.ON
                and new_color == self.state.color
            ):
                logger.debug('New color is same as current color')
                return True
        else:
            logger.debug('RGB Values not provided')
            new_color = None
            if brightness is not None:
                if Validators.validate_zero_to_hundred(brightness):
                    brightness_update = int(brightness)
                else:
                    logger.warning('Invalid brightness value')
                    return False
                if (
                    self.state.device_status == DeviceStatus.ON
                    and brightness_update == self.state.brightness
                ):
                    logger.debug('Brightness already set to %s', brightness)
                    return True
                color_mode = 'white'
            else:
                logger.warning('Brightness and RGB values are not set')
                return False

        data = {
            'action': DeviceStatus.ON,
            'speed': 0,
            'brightness': brightness_update,
            'red': 0 if new_color is None else int(new_color.rgb.red),
            'green': 0 if new_color is None else int(new_color.rgb.green),
            'blue': 0 if new_color is None else int(new_color.rgb.blue),
            'colorMode': 'color' if new_color is not None else 'white',
        }
        r_dict = await self.call_bypassv2_api('setLightStatus', data)

        r = Helpers.process_dev_response(logger, 'set_status', self, r_dict)
        if r is None:
            return False

        if color_mode == 'color' and new_color is not None:
            self.state.color_mode = 'color'
            self.state.color = new_color
        elif brightness is not None:
            self.state.brightness = int(brightness_update)
            self.state.color_mode = 'white'

        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    @deprecated(
        'toggle() is deprecated, use toggle_switch(toggle: bool | None = None) instead'
    )
    async def toggle(self, status: str) -> bool:
        """Toggle switch of VeSync ESL100MC."""
        status_bool = status == DeviceStatus.ON
        return await self.toggle_switch(status_bool)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status == DeviceStatus.OFF
        data = {'id': 0, 'enabled': toggle}

        r_dict = await self.call_bypassv2_api('setSwitch', data)

        r = Helpers.process_dev_response(logger, 'toggle', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.ON if toggle else DeviceStatus.OFF
        return True


class VeSyncBulbESL100(BypassV1Mixin, VeSyncBulb):
    """Object to hold VeSync ESL100 light bulb.

    Device state is held in the `state` attribute, which is an instance of
    [BulbState][pyvesync.base_devices.bulb_base.BulbState]. The `state` attribute
    contains all settable states for the bulb.

    This bulb only has the dimmable feature. Inherits from
    [VeSyncBulb][pyvesync.devices.vesyncbulb.VeSyncBulb] and
    [VeSyncBaseToggleDevice][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseToggleDevice].

    Args:
        details (dict): Dictionary of bulb state details.
        manager (VeSync): Manager class used to make API calls
        feature_map (BulbMap): Device configuration map.

    Attributes:
        state (BulbState): Device state object
            Each device has a separate state base class in the base_devices module.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: BulbMap
    ) -> None:
        """Initialize Etekcity ESL100 Dimmable Bulb."""
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        method_dict = {
            'method': 'deviceDetail',
        }

        r_dict = await self.call_bypassv1_api(
            bulb_models.RequestESL100Detail, method_dict, 'deviceDetail', 'deviceDetail'
        )
        model = process_bypassv1_result(
            self, logger, 'get_details', r_dict, bulb_models.ResponseESL100DetailResult
        )

        if model is None:
            self.state.connection_status = ConnectionStatus.OFFLINE
            return
        self.state.brightness = model.brightness
        self.state.device_status = model.deviceStatus
        self.state.connection_status = model.connectionStatus

    @deprecated(
        'toggle() is deprecated, use toggle_switch(toggle: bool | None = None) instead'
    )
    async def toggle(self, status: str) -> bool:
        """Toggle switch of ESL100 bulb."""
        status_bool = status != DeviceStatus.ON
        return await self.toggle_switch(status_bool)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        status = DeviceStatus.ON if toggle else DeviceStatus.OFF
        method_dict = {
            'status': status,
        }
        r_dict = await self.call_bypassv1_api(
            bulb_models.RequestESL100Status,
            method_dict,
            'smartBulbPowerSwitchCtl',
            'smartBulbPowerSwitchCtl',
        )

        r = Helpers.process_dev_response(logger, 'toggle', self, r_dict)
        if r is None:
            return False

        self.state.device_status = status
        return True

    @deprecated('Use set_brightness() instead')
    async def set_status(self, brightness: int) -> bool:
        """Set brightness of dimmable bulb.

        Args:
            brightness (int): Brightness of bulb (0-100).

        Returns:
            bool: True if successful, False otherwise.
        """
        return await self.set_brightness(brightness=brightness)

    async def set_brightness(self, brightness: int) -> bool:
        if not self.supports_brightness:
            logger.warning('%s is not dimmable', self.device_name)
            return False
        if not Validators.validate_zero_to_hundred(brightness):
            logger.warning('Invalid brightness value')
            return False
        brightness_update = brightness
        if (
            self.state.device_status == DeviceStatus.ON
            and brightness_update == self.supports_brightness
        ):
            logger.debug('Device already in requested state')
            return True

        method_dict = {
            'brightNess': str(brightness_update),
            'status': 'on',
        }
        r_dict = await self.call_bypassv1_api(
            bulb_models.RequestESL100Brightness,
            method_dict,
            'smartBulbBrightnessCtl',
            'smartBulbBrightnessCtl',
        )

        r = Helpers.process_dev_response(logger, 'set_brightness', self, r_dict)
        if r is None:
            return False

        self.state.brightness = brightness_update
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> None:
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1GetTimer, {}, 'getTimers', 'timer/getTimers'
        )
        result_model = process_bypassv1_result(
            self, logger, 'get_timer', r_dict, TimerModels.ResultV1GetTimer
        )
        if result_model is None:
            return
        if not isinstance(result_model.timers, list) or not result_model.timers:
            self.state.timer = None
            logger.debug('No timers found')
            return
        timer = result_model.timers
        if not isinstance(timer, TimerModels.TimerItemV1):
            logger.warning('Invalid timer item type')
            return
        self.state.timer = Timer(
            int(timer.counterTimer),
            timer.action,
            int(timer.timerID),
        )

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            logger.warning("Invalid action value - must be 'on' or 'off'")
            return False
        update_dict = {
            'action': action,
            'counterTime': str(duration),
            'status': '1',
        }
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1SetTime, update_dict, 'addTimer', 'timer/addTimer'
        )
        result_model = process_bypassv1_result(
            self, logger, 'set_timer', r_dict, TimerModels.ResultV1SetTimer
        )
        if result_model is None:
            return False
        self.state.timer = Timer(duration, action, int(result_model.timerID))
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer set - run get_timer() first')
            return False
        timer = self.state.timer
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1ClearTimer,
            {'timerId': str(timer.id), 'status': '1'},
            'deleteTimer',
            'timer/deleteTimer',
        )
        result = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if result is None:
            return False
        self.state.timer = None
        return True


class VeSyncBulbESL100CW(BypassV1Mixin, VeSyncBulb):
    """VeSync Tunable and Dimmable White Bulb.

    This bulb only has the dimmable feature. Inherits from
    [VeSyncBulb][pyvesync.devices.vesyncbulb.VeSyncBulb] and
    [VeSyncBaseToggleDevice][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseToggleDevice].

    Device state is held in the `state` attribute, which is an instance of
    [BulbState][pyvesync.base_devices.bulb_base.BulbState]. The `state` attribute
    contains all settable states for the bulb.

    Args:
        details (dict): Dictionary of bulb state details.
        manager (VeSync): Manager class used to make API calls
        feature_map (BulbMap): Device configuration map.

    Attributes:
        state (BulbState): Device state object
            Each device has a separate state base class in the base_devices module.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: BulbMap
    ) -> None:
        """Initialize Etekcity Tunable white bulb."""
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv1_api(
            bulb_models.RequestESL100CWBase,
            {'jsonCmd': {'getLightStatus': 'get'}},
        )

        light_resp = process_bypassv1_result(
            self, logger, 'get_details', r_dict, bulb_models.ResponseESL100CWDetailResult
        )
        if light_resp is None:
            self.state.connection_status = ConnectionStatus.OFFLINE
            return
        self._interpret_apicall_result(light_resp)

    def _interpret_apicall_result(
        self, response: bulb_models.ResponseESL100CWDetailResult
    ) -> None:
        self.state.connection_status = ConnectionStatus.ONLINE
        result = response.light
        self.state.device_status = result.action
        self.state.brightness = result.brightness
        self.state.color_temp = result.colorTempe

    @deprecated(
        'toggle() is deprecated, use toggle_switch(toggle: bool | None = None) instead'
    )
    async def toggle(self, status: str) -> bool:
        """Deprecated - use toggle_switch()."""
        status_bool = status == DeviceStatus.ON
        return await self.toggle_switch(status_bool)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status == DeviceStatus.OFF
        status = DeviceStatus.ON if toggle else DeviceStatus.OFF

        r_dict = await self.call_bypassv1_api(
            bulb_models.RequestESL100CWBase,
            {'jsonCmd': {'light': {'action': status}}},
            'bypass',
            'bypass',
        )

        r = Helpers.process_dev_response(logger, 'toggle', self, r_dict)
        if r is None:
            logger.info('%s offline', self.device_name)
            return False
        self.state.device_status = status
        return True

    async def set_brightness(self, brightness: int) -> bool:
        """Set brightness of tunable bulb."""
        return await self.set_status(brightness=brightness)

    async def set_status(
        self, /, brightness: int | None = None, color_temp: int | None = None
    ) -> bool:
        """Set status of tunable bulb."""
        if brightness is not None:
            if Validators.validate_zero_to_hundred(brightness):
                brightness_update = int(brightness)
            else:
                logger.warning('Invalid brightness value')
                return False
        elif self.state.brightness is not None:
            brightness_update = self.state.brightness
        else:
            brightness_update = 100
        if color_temp is not None:
            if Validators.validate_zero_to_hundred(color_temp):
                color_temp_update = color_temp
            else:
                logger.warning('Invalid color temperature value')
                return False
        elif self.state.color_temp is not None:
            color_temp_update = self.state.color_temp
        else:
            color_temp_update = 100
        if (
            self.state.device_status == DeviceStatus.ON
            and brightness_update == self.state.brightness
            and color_temp_update == self.state.color_temp
        ):
            logger.debug('Device already in requested state')
            return True
        light_dict: dict[str, NUMERIC_OPT | str] = {
            'colorTempe': color_temp_update,
            'brightness': brightness_update,
            'action': DeviceStatus.ON,
        }

        r_dict = await self.call_bypassv1_api(
            bulb_models.RequestESL100CWBase,
            {'jsonCmd': {'light': light_dict}},
            'bypass',
            'bypass',
        )

        r = Helpers.process_dev_response(logger, 'set_brightness', self, r_dict)
        if r is None:
            return False

        self.state.brightness = brightness_update
        self.state.color_temp = color_temp_update
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_color_temp(self, color_temp: int) -> bool:
        return await self.set_status(color_temp=color_temp)

    async def get_timer(self) -> None:
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1GetTimer, {}, 'getTimers', 'timer/getTimers'
        )
        result_model = process_bypassv1_result(
            self, logger, 'get_timer', r_dict, TimerModels.ResultV1GetTimer
        )
        if result_model is None:
            return
        if not isinstance(result_model.timers, list) or not result_model.timers:
            logger.debug('No timers found')
            return
        timers = result_model.timers
        if len(timers) > 1:
            logger.debug('Multiple timers found, returning first timer')
        timer = timers[0]
        if not isinstance(timer, TimerModels.TimeItemV1):
            logger.warning('Invalid timer item type')
            return
        self.state.timer = Timer(
            int(timer.counterTime),
            timer.action,
            int(timer.timerID),
        )

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            logger.warning("Invalid action value - must be 'on' or 'off'")
            return False
        update_dict = {
            'action': action,
            'counterTime': str(duration),
            'status': '1',
        }
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1SetTime, update_dict, 'addTimer', 'timer/addTimer'
        )
        result_model = process_bypassv1_result(
            self, logger, 'set_timer', r_dict, TimerModels.ResultV1SetTimer
        )
        if result_model is None:
            return False
        self.state.timer = Timer(duration, action, int(result_model.timerID))
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer set - run get_timer() first')
            return False
        timer = self.state.timer
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1ClearTimer,
            {'timerId': str(timer.id), 'status': '1'},
            'deleteTimer',
            'timer/deleteTimer',
        )
        r = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if r is None:
            return False
        self.state.timer = None
        return True


class VeSyncBulbValcenoA19MC(VeSyncBulb):
    """VeSync Multicolor Bulb.

    This bulb only has the dimmable feature. Inherits from
    [VeSyncBulb][pyvesync.devices.vesyncbulb.VeSyncBulb] and
    [VeSyncBaseToggleDevice][pyvesync.base_devices.vesyncbasedevice.VeSyncBaseToggleDevice].

    Device state is held in the `state` attribute, which is an instance of
    [BulbState][pyvesync.base_devices.bulb_base.BulbState]. The `state` attribute
    contains all settable states for the bulb.

    Args:
        details (dict): Dictionary of bulb state details.
        manager (VeSync): Manager class used to make API calls
        feature_map (BulbMap): Device configuration map.

    Attributes:
        state (BulbState): Device state object
            Each device has a separate state base class in the base_devices module.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: BulbMap
    ) -> None:
        """Initialize Multicolor bulb."""
        super().__init__(details, manager, feature_map)
        self.request_keys = (
            'acceptLanguage',
            'accountID',
            'appVersion',
            'cid',
            'configModule',
            'deviceRegion',
            'debugMode',
            'phoneBrand',
            'phoneOS',
            'timeZone',
            'token',
            'traceId',
        )

    def _payload_base(self) -> bulb_models.ValcenoStatusPayload:
        """Return the payload base for the set status request.

        Avoid duplicating code and ensure that the payload dict is
        reset to its default state before each request.
        """
        payload_dict: bulb_models.ValcenoStatusPayload = {
            'force': 0,
            'brightness': '',
            'colorTemp': '',
            'colorMode': '',
            'hue': '',
            'saturation': '',
            'value': '',
        }
        return payload_dict

    def _build_request(self, payload: dict) -> dict:
        """Build request for Valceno Smart Bulb.

        The request is built from the keys in `self.request_keys` and the
        `payload` dict. The payload is then added to the request body.

        Args:
            method (str): The method to use for the request.
            payload (dict): The data to use for the payload.

        Returns:
            dict: The request body.

        Note:
            The `payload` argument is the value of request_body['payload'].
            The request structure is:
            ```json
            {
                "acceptLanguage": "en",
                "accountID": "1234567890",
                "appVersion": "1.0.0",
                "cid": "1234567890",
                "configModule": "default",
                "deviceRegion": "US",
                "debugMode": False,
                "phoneBrand": "Apple",
                "phoneOS": "iOS",
                "timeZone": "New_York",
                "token": "",
                "traceId": "",
                "payload": {
                    "method": "getLightStatusV2",
                    "source": "APP",
                    "data": {}
                }
            }
            ```
        """
        default_dict = Helpers.get_defaultvalues_attributes(self.request_keys)
        default_dict.update(Helpers.get_device_attributes(self, self.request_keys))
        default_dict.update(
            Helpers.get_manager_attributes(self.manager, self.request_keys)
        )
        default_dict['method'] = 'bypassV2'
        default_dict['payload'] = payload
        return default_dict

    async def _call_valceno_api(
        self, payload_method: str, payload_data: Mapping
    ) -> dict | None:
        """Make API call to Valceno Smart Bulb.

        Args:
            payload_method (str): The method to use for the request.
            payload_data (dict): The payload to use for the request.

        Returns:
            tuple[bytes, Any]: The response from the API call.

        Note:
            The request is built by the `_build_request` method and the payload is
            added to the request body.
            The payload structure is:
            ```json
            {
                "method": "getLightStatusV2",
                "source": "APP",
                "data": {}
                }
        """
        payload = {'method': payload_method, 'source': 'APP', 'data': payload_data}
        request_body = self._build_request(payload)
        r_dict, _ = await self.manager.async_call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=Helpers.req_header_bypass(),
            json_object=request_body,
        )
        resp = Helpers.process_dev_response(logger, payload_method, self, r_dict)
        if resp is None:
            return None
        return r_dict

    async def get_details(self) -> None:
        r_dict = await self._call_valceno_api('getLightStatusV2', {})

        if r_dict is None:
            return

        status = bulb_models.ResponseValcenoStatus.from_dict(r_dict)
        self._interpret_apicall_result(status)

    def _interpret_apicall_result(
        self, response: bulb_models.ResponseValcenoStatus
    ) -> None:
        """Process API response with device status.

        Assigns the values to the state attributes of the device.

        Args:
            response (bulb_models.ResponseValcenoStatus): The response from the API call.
        """
        result = response.result.result
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = result.enabled
        self.state.brightness = result.brightness
        self.state.color_temp = result.colorTemp
        self.state.color_mode = result.colorMode
        hue = float(round(result.hue / 250 * 9, 2))
        sat = float(result.saturation / 100)
        val = float(result.value)
        self.state.color = Color.from_hsv(hue=hue, saturation=sat, value=val)

    @deprecated(
        'toggle() is deprecated, use toggle_switch(toggle: bool | None = None) instead'
    )
    async def toggle(self, status: str) -> bool:
        """Deprecated - use toggle_switch()."""
        status_bool = status == DeviceStatus.ON
        return await self.toggle_switch(status_bool)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status == DeviceStatus.OFF

        if toggle == self.state.device_status:
            logger.warning('Device already in requested state')
            return True

        payload_data = {
            'id': 0,
            'enabled': toggle,
        }
        method = 'setSwitch'

        r_dict = await self._call_valceno_api(method, payload_data)

        if r_dict is None:
            self.state.device_status = DeviceStatus.OFF
            return False
        self.state.device_status = DeviceStatus.ON if toggle else DeviceStatus.OFF
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_rgb(self, red: float, green: float, blue: float) -> bool:
        new_color = Color.from_rgb(red=red, green=green, blue=blue)
        if new_color is None:
            logger.warning('Invalid RGB values')
            return False

        return await self.set_hsv(
            hue=new_color.hsv.hue,
            saturation=new_color.hsv.saturation,
            value=new_color.hsv.value,
        )

    async def set_brightness(self, brightness: int) -> bool:
        """Set brightness of multicolor bulb."""
        return await self.set_status(brightness=brightness)

    async def set_color_temp(self, color_temp: int) -> bool:
        """Set White Temperature of Bulb in pct (0 - 100)."""
        return await self.set_status(color_temp=color_temp)

    async def set_color_hue(self, color_hue: float) -> bool:
        """Set Color Hue of Bulb (0 - 360)."""
        return await self.set_status(color_hue=color_hue)

    async def set_color_saturation(self, color_saturation: float) -> bool:
        """Set Color Saturation of Bulb in pct (1 - 100)."""
        return await self.set_status(color_saturation=color_saturation)

    async def set_color_value(self, color_value: float) -> bool:
        """Set Value of multicolor bulb in pct (1 - 100)."""
        # Equivalent to brightness level, when in color mode.
        return await self.set_status(color_value=color_value)

    async def set_color_mode(self, color_mode: str) -> bool:
        """Set Color Mode of Bulb (white / hsv)."""
        return await self.set_status(color_mode=color_mode)

    async def set_hsv(self, hue: float, saturation: float, value: float) -> bool:
        new_color = Color.from_hsv(hue=hue, saturation=saturation, value=value)
        if new_color is None:
            logger.warning('Invalid HSV values')
            return False

        # the api expects the hsv Value in the brightness parameter
        payload_data = self._build_status_payload(
            hue=hue,
            saturation=saturation,
            value=value,
        )
        if payload_data is None:
            return False
        resp = await self._call_valceno_api('setLightStatusV2', payload_data)
        if resp is None:
            return False

        r_dict = Helpers.process_dev_response(logger, 'set_hsv', self, resp)
        if r_dict is None:
            return False

        status = bulb_models.ResponseValcenoStatus.from_dict(r_dict)
        self._interpret_apicall_result(status)
        return True

    async def set_white_mode(self) -> bool:
        return await self.set_status(color_mode='white')

    async def set_status(
        self,
        *,
        brightness: float | None = None,
        color_temp: float | None = None,
        color_hue: float | None = None,
        color_saturation: float | None = None,
        color_value: float | None = None,
        color_mode: str | None = None,
    ) -> bool:
        """Set multicolor bulb parameters.

        No arguments turns bulb on. **Kwargs only**

        Args:
            brightness (int, optional): brightness between 0 and 100
            color_temp (int, optional): color temperature between 0 and 100
            color_mode (int, optional): color mode hsv or white
            color_hue (float, optional): color hue between 0 and 360
            color_saturation (float, optional): color saturation between 0 and 100
            color_value (int, optional): color value between 0 and 100

        Returns:
            bool : True if call was successful, False otherwise
        """
        payload_data = self._build_status_payload(
            brightness=brightness,
            color_temp=color_temp,
            hue=color_hue,
            saturation=color_saturation,
            value=color_value,
            color_mode=color_mode,
        )
        if payload_data == self._payload_base():
            logger.debug('No state change.')
            return False
        if payload_data is None:
            logger.warning(
                'Invalid payload data for set_status call for %s', self.device_name
            )
            return False

        r_dict = await self._call_valceno_api('setLightStatusV2', payload_data)
        if r_dict is None:
            return False

        r_dict = Helpers.process_dev_response(logger, 'set_status', self, r_dict)
        if r_dict is None:
            return False

        r_model = bulb_models.ResponseValcenoStatus.from_dict(r_dict)
        self._interpret_apicall_result(r_model)
        return True

    def _build_color_payload(
        self,
        color_hue: float | None = None,
        color_saturation: float | None = None,
        color_value: float | None = None,
    ) -> bulb_models.ValcenoStatusPayload | None:
        """Create color payload for Valceno Bulbs.

        This is called by `_build_status_payload` if any of the HSV values are set.
        This should not be called directly.

        Args:
            color_hue (float, optional): Color hue of bulb (0-360).
            color_saturation (float, optional): Color saturation of bulb (0-100).
            color_value (float, optional): Color value of bulb (0-100).
        """
        payload_dict = self._payload_base()
        new_color = Color.from_hsv(
            hue=color_hue, saturation=color_saturation, value=color_value
        )
        if new_color is None:
            logger.warning('Invalid HSV values')
            return None
        payload_dict['hue'] = int(new_color.hsv.hue * 250 / 9)
        payload_dict['saturation'] = int(new_color.hsv.saturation * 100)
        payload_dict['value'] = int(new_color.hsv.value)
        payload_dict['colorMode'] = 'hsv'
        payload_dict['force'] = 1
        return payload_dict

    def _build_status_payload(
        self,
        brightness: float | None = None,
        color_temp: float | None = None,
        hue: float | None = None,
        saturation: float | None = None,
        value: float | None = None,
        color_mode: str | None = None,
    ) -> bulb_models.ValcenoStatusPayload | None:
        """Create status payload data for Valceno Bulbs.

        If color_mode is set, hue, saturation and/or value must be set as well.
        If color_mode is not set, brightness and/or color_temp must be set.

        This builds the `request_body['payload']['data']` dict for api calls that
        set the status of the bulb.

        Args:
            brightness (float, optional): Brightness of bulb (0-100).
            color_temp (float, optional): Color temperature of bulb (0-100).
            hue (float, optional): Color hue of bulb (0-360).
            saturation (float, optional): Color saturation of bulb (0-100).
            value (float, optional): Color value of bulb (0-100).
            color_mode (str, optional): Color mode of bulb ('white', 'hsv', 'color').

        Returns:
            dict | None: Payload dictionary to be sent to the API.
        """
        payload_dict = self._payload_base()
        if all(itm is not None for itm in [hue, saturation, value]):
            color_dict = self._build_color_payload(
                color_hue=hue, color_saturation=saturation, color_value=value
            )
            if color_dict is not None:
                return color_dict
            # Return None if HSV values are same as current state
            if self._check_color_state(hue=hue, saturation=saturation, value=value):
                return None
        elif color_mode in ['color', 'hsv', 'rgb']:
            logger.debug('HSV values must be provided when setting color mode.')
        else:
            if color_mode == 'white' and not Validators.validate_zero_to_hundred(
                color_temp
            ):
                payload_dict['colorMode'] = 'white'
            if color_temp is not None and Validators.validate_zero_to_hundred(color_temp):
                payload_dict['colorTemp'] = int(color_temp)
                payload_dict['colorMode'] = 'white'
            if brightness is not None and Validators.validate_zero_to_hundred(brightness):
                payload_dict['brightness'] = int(brightness)
        force_keys = ['colorTemp', 'saturation', 'hue', 'colorMode', 'value']
        for key in force_keys:
            if payload_dict.get(key) != '':
                payload_dict['force'] = 1
        return payload_dict

    def _check_color_state(
        self,
        hue: float | None = None,
        saturation: float | None = None,
        value: float | None = None,
    ) -> bool:
        """Check if color state is already set.

        Returns True if color is already set, False otherwise.

        Args:
            hue (float, optional): Color hue.
            saturation (float, optional): Color saturation.
            value (float, optional): Color value.
        """
        if self.state.color is not None:
            set_hue = hue or self.state.color.hsv.hue
            set_saturation = saturation or self.state.color.hsv.saturation
            set_value = value or self.state.color.hsv.value
            set_color = Color.from_hsv(
                hue=set_hue, saturation=set_saturation, value=set_value
            )
            if (
                self.state.device_status == DeviceStatus.ON
                and set_color == self.state.color
            ):
                logger.debug('Device already set to requested color')
                return True
        return False



"""
This tests all requests made by the pyvesync library with pytest.

All tests inherit from the TestBase class which contains the fixtures
and methods needed to run the tests.

The `helpers.call_api` method is patched to return a mock response.
The method, endpoint, headers and json arguments are recorded
in YAML files in the api directory, categorized in folders by
module and files by the class name.

The default is to record requests that do not exist and compare requests
that already exist. If the API changes, set the overwrite argument to True
in order to overwrite the existing YAML file with the new request.
"""
import logging
from base_test_cases import TestBase
import call_json
import call_json_outlets
import call_json_bulbs
import call_json_fans
import call_json_purifiers
import call_json_humidifiers
import call_json_switches
from utils import assert_test, parse_args

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_device_tests():
    """Test to ensure all devices have are defined for testing.

    All devices should have an entry in the DETAILS_RESPONSES dict
    with response for get_details() method. This test ensures that
    all devices have been configured for testing. The details response
    should be located in `{device_type}Details` class of the respective
    call_json_{device_type} module and the DETAILS_RESPONSE module variable.
    The class variable with the details response does not matter, the dictionary
    key of DETAILS_RESPONSES should match the device type.

    Examples
    ---------
    class FanDetails:
        "Core200SResponse": {'speed': 1, 'device_status': 'on'}

    DETAILS_RESPONSES = {
        'Core200S': FanDetails.Core200SResponse
    }

    Asserts
    -------
    Number of devices for each type has a response defined in the
        respective `call_json` module.

    See Also
    --------
    src/tests/README.md - README located in the tests directory
    """
    assert call_json_fans.FANS_NUM == len(call_json_fans.DETAILS_RESPONSES)
    assert call_json_bulbs.BULBS_NUM == len(call_json_bulbs.DETAILS_RESPONSES)
    assert call_json_outlets.OUTLETS_NUM == len(call_json_outlets.DETAILS_RESPONSES)
    assert call_json_switches.SWITCHES_NUM == len(call_json_switches.DETAILS_RESPONSES)
    assert call_json_purifiers.PURIFIERS_NUM == len(call_json_purifiers.DETAILS_RESPONSES)
    assert call_json_humidifiers.HUMIDIFIERS_NUM == len(call_json_humidifiers.DETAILS_RESPONSES)


class TestGeneralAPI(TestBase):
    """General API testing class for login() and get_devices()."""

    def test_get_devices(self):
        """Test get_devices() method request and API response."""
        print("Test Device List")
        self.mock_api.return_value = call_json.DeviceList.device_list_response(), 200
        self.run_in_loop(self.manager.get_devices)
        all_kwargs = parse_args(self.mock_api)
        assert assert_test(self.manager.get_devices, all_kwargs, None,
                           self.write_api, self.overwrite)
        assert len(self.manager.devices.bulbs) == call_json_bulbs.BULBS_NUM
        assert len(self.manager.devices.outlets) == call_json_outlets.OUTLETS_NUM
        assert len(self.manager.devices.fans) == call_json_fans.FANS_NUM
        assert len(self.manager.devices.switches) == call_json_switches.SWITCHES_NUM
        assert len(self.manager.devices.humidifiers) == call_json_humidifiers.HUMIDIFIERS_NUM
        assert len(self.manager.devices.air_purifiers) == call_json_purifiers.PURIFIERS_NUM



"""Etekcity 7A Outlet tests."""

import orjson
from pyvesync.devices.vesyncoutlet import VeSyncOutlet7A
from pyvesync.utils.helpers import Helpers as Helpers
import call_json
import call_json_outlets
from base_test_cases import TestBase

DEVICE_TYPE = 'wifi-switch-1.3'

DEV_LIST_DETAIL = call_json.DeviceList.device_list_item(DEVICE_TYPE)

CORRECT_7A_LIST = call_json.DeviceList.device_list_response(DEVICE_TYPE)

CORRECT_7A_DETAILS = call_json_outlets.DETAILS_RESPONSES[DEVICE_TYPE]

ENERGY_HISTORY = call_json_outlets.ENERGY_HISTORY

DEVICE_TYPE = 'wifi-switch-1.3'

CALL_LIST = [
    'turn_on',
    'turn_off',
    'update'
]


class TestVesync7ASwitch(TestBase):
    """Test 7A outlet API."""

    def test_7aswitch_conf(self):
        """Test initialization of 7A outlet."""
        self.mock_api.return_value = CORRECT_7A_LIST
        self.run_in_loop(self.manager.get_devices)
        outlets = self.manager.outlets
        assert len(outlets) == 1
        vswitch7a = outlets[0]
        assert isinstance(vswitch7a, VeSyncOutlet7A)
        assert vswitch7a.device_name == call_json.TestDefaults.name(DEVICE_TYPE)
        assert vswitch7a.device_type == DEVICE_TYPE
        assert vswitch7a.cid == call_json.TestDefaults.cid(DEVICE_TYPE)
        assert vswitch7a.is_on

    def test_7a_details(self):
        """Test get_details() method for 7A outlet."""
        resp_dict, status = CORRECT_7A_DETAILS
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch7a.get_details)
        dev_details = vswitch7a.details
        assert vswitch7a.device_status == 'on'
        assert isinstance(dev_details, dict)
        assert dev_details['active_time'] == 1
        assert dev_details['energy'] == 1
        assert vswitch7a.power == 1
        assert vswitch7a.voltage == 1

    def NO_test_7a_no_devstatus(self):
        """Test 7A outlet details response with no device status key."""
        # Remove test for invalid key - will be addressed by Validation
        bad_7a_details = {
            'deviceImg': '',
            'activeTime': 1,
            'energy': 1,
            'power': '1A:1A',
            'voltage': '1A:1A',
        }
        self.mock_api.return_value = (orjson.dumps(bad_7a_details), 200)
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch7a.get_details)
        assert len(self.caplog.records) == 1
        assert 'details' in self.caplog.text

    def NO_test_7a_no_details(self):
        """Test 7A outlet details response with unknown keys."""
        # Remove test for invalid key - will be addressed by Validation
        bad_7a_details = {'wrongdetails': 'on'}
        self.mock_api.return_value = (orjson.dumps(bad_7a_details), 200)
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch7a.get_details)
        assert len(self.caplog.records) == 1

    def test_7a_onoff(self):
        """Test 7A outlet on/off methods."""
        self.mock_api.return_value = (None, 200)
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        on = self.run_in_loop(vswitch7a.turn_on)
        head = Helpers.req_legacy_headers(self.manager)
        self.mock_api.assert_called_with(
            '/v1/wifi-switch-1.3/' + vswitch7a.cid + '/status/on', 'put', headers=head
        )
        assert on
        off = self.run_in_loop(vswitch7a.turn_off)
        self.mock_api.assert_called_with(
            '/v1/wifi-switch-1.3/' + vswitch7a.cid + '/status/off', 'put', headers=head
        )
        assert off

    def test_7a_onoff_fail(self):
        """Test 7A outlet on/off methods that fail."""
        self.mock_api.return_value = (None, 400)
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        assert not self.run_in_loop(vswitch7a.turn_on)
        assert not self.run_in_loop(vswitch7a.turn_off)

    def test_7a_weekly(self):
        """Test 7A outlet weekly energy API call and energy dict."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch7a.get_weekly_energy)
        self.mock_api.assert_called_with(
            '/v1/device/' + vswitch7a.cid + '/energy/week',
            'get',
            headers=Helpers.req_legacy_headers(self.manager),
        )
        energy_dict = vswitch7a.energy['week']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]

    def test_7a_monthly(self):
        """Test 7A outlet monthly energy API call and energy dict."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch7a.get_monthly_energy)
        self.mock_api.assert_called_with(
            '/v1/device/' + vswitch7a.cid + '/energy/month',
            'get',
            headers=Helpers.req_legacy_headers(self.manager),
        )
        energy_dict = vswitch7a.energy['month']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]

    def test_7a_yearly(self):
        """Test 7A outlet yearly energy API call and energy dict."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch7a.get_yearly_energy)
        self.mock_api.assert_called_with(
            '/v1/device/' + vswitch7a.cid + '/energy/year',
            'get',
            headers=Helpers.req_legacy_headers(self.manager),
        )
        energy_dict = vswitch7a.energy['year']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]

    def test_history_fail(self):
        """Test handling of energy update failure."""
        bad_history = {'code': 1}
        self.mock_api.return_value = (orjson.dumps(bad_history), 200)
        vswitch7a = VeSyncOutlet7A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch7a.update_energy)
        assert len(self.caplog.records) == 1
        assert 'weekly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(vswitch7a.get_monthly_energy)
        assert len(self.caplog.records) == 1
        assert 'monthly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(vswitch7a.get_yearly_energy)
        assert len(self.caplog.records) == 1
        assert 'yearly' in self.caplog.text



"""Models for general VeSync API requests and responses.

Dataclasses should follow the naming convention of Request/Response + <API Name> + Model.
Internal models should be named starting with IntResp/IntReq<API Name>Model.

Note:
    All models should inherit `ResponseBaseModel` or `RequestBaseModel`. Use
    `pyvesync.models.base_models.DefaultValues` to set default values. There
    should be no repeating keys set in the child models.

"""

from __future__ import annotations

import hashlib
from dataclasses import dataclass, field
from typing import Any

from pyvesync.models.base_models import (
    DefaultValues,
    RequestBaseModel,
    ResponseBaseModel,
    ResponseCodeModel,
)


@dataclass
class RequestGetTokenModel(RequestBaseModel):
    """Request model for requesting auth token (used for login)."""

    # Arguments to set
    email: str
    method: str
    password: str
    # default values
    acceptLanguage: str = DefaultValues.acceptLanguage
    accountID: str = ''
    authProtocolType: str = 'generic'
    clientInfo: str = DefaultValues.phoneBrand
    clientType: str = DefaultValues.clientType
    clientVersion: str = f'VeSync {DefaultValues.appVersion}'
    debugMode: bool = False
    osInfo: str = DefaultValues.phoneOS
    terminalId: str = DefaultValues.terminalId
    timeZone: str = DefaultValues.timeZone
    token: str = ''
    userCountryCode: str = DefaultValues.userCountryCode
    appID: str = DefaultValues.appId
    sourceAppID: str = DefaultValues.appId
    traceId: str = field(default_factory=DefaultValues.newTraceId)

    def __post_init__(self) -> None:
        """Hash the password field."""
        self.password = self.hash_password(self.password)

    @staticmethod
    def hash_password(string: str) -> str:
        """Encode password."""
        return hashlib.md5(string.encode('utf-8')).hexdigest()  # noqa: S324


@dataclass
class RespGetTokenResultModel(ResponseBaseModel):
    """Model for the 'result' field in auth response with authorizeCode and account ID.

    This class is referenced by the `ResponseAuthModel` class.
    """

    accountID: str
    authorizeCode: str


@dataclass
class RequestLoginTokenModel(RequestBaseModel):
    """Request model for login."""

    # Arguments to set
    method: str
    authorizeCode: str | None
    # default values
    acceptLanguage: str = DefaultValues.acceptLanguage
    accountID: str = ''
    clientInfo: str = DefaultValues.phoneBrand
    clientType: str = DefaultValues.clientType
    clientVersion: str = f'VeSync {DefaultValues.appVersion}'
    debugMode: bool = False
    emailSubscriptions: bool = False
    osInfo: str = DefaultValues.phoneOS
    terminalId: str = DefaultValues.terminalId
    timeZone: str = DefaultValues.timeZone
    token: str = ''
    bizToken: str | None = None
    regionChange: str | None = None
    userCountryCode: str = DefaultValues.userCountryCode
    traceId: str = field(default_factory=DefaultValues.newTraceId)

    def __post_serialize__(self, d: dict[Any, Any]) -> dict[Any, Any]:
        """Remove null keys."""
        if d['regionChange'] is None:
            d.pop('regionChange')
        if d['authorizeCode'] is None:
            d.pop('authorizeCode')
        if d['bizToken'] is None:
            d.pop('bizToken')
        return d


@dataclass
class RespLoginTokenResultModel(ResponseBaseModel):
    """Model for the 'result' field in login response containing token and account ID.

    This class is referenced by the `ResponseLoginModel` class.
    """

    accountID: str
    acceptLanguage: str
    countryCode: str
    token: str
    bizToken: str = ''
    currentRegion: str = ''


@dataclass
class ResponseLoginModel(ResponseCodeModel):
    """Model for the login response.

    Inherits from `BaseResultModel`. The `BaseResultModel` class provides the
    defaults "code" and "msg" fields for the response.

    Attributes:
        result: ResponseLoginResultModel
            The inner model for the 'result' field in the login response.

    Examples:
        ```python
        a = {
            "code": 0,
            "msg": "success",
            "stacktrace": null,
            "module": null,
            "traceId": "123456",
            "result": {
                "accountID": "123456",
                "acceptLanguage": "en",
                "countryCode": "US",
                "token": "abcdef1234567890"
            }
        }
        b = ResponseLoginModel.from_dict(a)
        account_id = b.result.accountId
        token = b.result.token
        ```
    """

    result: RespLoginTokenResultModel | RespGetTokenResultModel


@dataclass
class RequestDeviceListModel(RequestBaseModel):
    """Model for the device list request."""

    token: str
    accountID: str
    timeZone: str = DefaultValues.timeZone
    method: str = 'devices'
    pageNo: int = 1
    pageSize: int = 100
    appVersion: str = DefaultValues.appVersion
    phoneBrand: str = DefaultValues.phoneBrand
    phoneOS: str = DefaultValues.phoneOS
    acceptLanguage: str = DefaultValues.acceptLanguage
    traceId: str = str(DefaultValues.traceId())


def _flatten_device_prop(d: dict[str, Any]) -> dict[str, Any]:
    """Flatten deviceProp field in device list response."""
    if isinstance(d.get('deviceProp'), dict):
        device_prop = d['deviceProp']
        if device_prop.get('powerSwitch') is not None:
            d['deviceStatus'] = 'on' if device_prop['powerSwitch'] == 1 else 'off'
        if device_prop.get('connectionStatus') is not None:
            d['connectionStatus'] = device_prop['connectionStatus']
        if device_prop.get('wifiMac') is not None:
            d['macID'] = device_prop['wifiMac']
    d.pop('deviceProp', None)
    return d


@dataclass
class ResponseDeviceDetailsModel(ResponseBaseModel):
    """Internal response model for each device in device list response.

    Populates the 'list' field in the `InternalDeviceListResult`.

    Certain Devices have device status information in the `deviceProp` or `extension`
    fields. This model flattens those fields into the `deviceStatus` and
    `connectionStatus` fields before deserialization.
    """

    deviceRegion: str
    isOwner: bool
    deviceName: str
    cid: str
    connectionType: str
    deviceType: str
    type: str
    configModule: str
    uuid: str | None = None
    macID: str = ''
    mode: str = ''
    deviceImg: str = ''
    speed: str | None = None
    currentFirmVersion: str | None = None
    subDeviceType: str | None = None
    subDeviceList: str | None = None
    extension: InternalDeviceListExtension | None = None
    subDeviceNo: int | None = None
    deviceStatus: str = 'off'
    connectionStatus: str = 'offline'
    productType: str | None = None

    @classmethod
    def __pre_deserialize__(cls, d: dict[Any, Any]) -> dict[Any, Any]:
        """Perform device_list pre-deserialization processes.

        This performs the following:
            - Flattens the `deviceProp` field into `deviceStatus`, `connectionStatus`
                and `macID`
            - Sets `cid` to `uuid` or `macID` if null
        """
        super().__pre_deserialize__(d)
        d = _flatten_device_prop(d)
        if d.get('cid') is None:
            d['cid'] = d.get('uuid') if d.get('uuid') is not None else d.get('macID')
        return d


@dataclass
class InternalDeviceListResult(ResponseBaseModel):
    """Internal model for the 'result' field in device list response.

    Notes:
      Used by the `ResponseDeviceListModel` class to populate result field.
    """

    total: int
    pageSize: int
    pageNo: int
    list: list[ResponseDeviceDetailsModel]


@dataclass
class ResponseDeviceListModel(ResponseCodeModel):
    """Device list response model.

    Inherits from `BaseResultModel`. The `BaseResultModel` class provides the
    defaults "code" and "msg" fields for the response.

    Attributes:
        result: InternalDeviceListResult
            The inner model for the 'result' field in the device list response.
        module: str | None
        stacktrace: str | None

    Notes:
        See the `DeviceListResultModel` and `DeviceListDeviceModel` classes for
        the inner model of the 'result' field.
    """

    result: InternalDeviceListResult


@dataclass
class InternalDeviceListExtension(ResponseBaseModel):
    """Internal Optional 'extension' field in device list response.

    Used by the `InnerRespDeviceListDevModel` class to populate
    the extension field in the device list response.
    """

    airQuality: None | int
    airQualityLevel: None | int
    mode: None | str
    fanSpeedLevel: None | str


@dataclass
class RequestPID(RequestBaseModel):
    """Model for the PID request."""

    method: str
    appVersion: str
    phoneBrand: str
    phoneOS: str
    traceId: str
    token: str
    accountID: str
    mobileID: str
    configModule: str
    region: str


@dataclass
class RequestFirmwareModel(RequestBaseModel):
    """Model for the firmware request."""

    accountID: str
    timeZone: str
    token: str
    userCountryCode: str
    cidList: list[str]
    acceptLanguage: str = DefaultValues.acceptLanguage
    traceId: str = field(default_factory=DefaultValues.traceId)
    appVersion: str = DefaultValues.appVersion
    phoneBrand: str = DefaultValues.phoneBrand
    phoneOS: str = DefaultValues.phoneOS
    method: str = 'getFirmwareUpdateInfoList'
    debugMode: bool = False


@dataclass
class ResponseFirmwareModel(ResponseCodeModel):
    """Model for the firmware response."""

    result: FirmwareResultModel


@dataclass
class FirmwareUpdateInfoModel(ResponseBaseModel):
    """Firmware update information model."""

    currentVersion: str
    latestVersion: str
    releaseNotes: str
    pluginName: str
    isMainFw: bool


@dataclass
class FirmwareDeviceItemModel(ResponseBaseModel):
    """Model for the firmware device item in the firmware response."""

    deviceCid: str
    deviceName: str
    code: int
    msg: str | None
    firmUpdateInfos: list[FirmwareUpdateInfoModel]


@dataclass
class FirmwareResultModel(ResponseBaseModel):
    """Model for the firmware response result."""

    cidFwInfoList: list[FirmwareDeviceItemModel]


@dataclass
class RequestDeviceConfiguration(RequestBaseModel):
    """Model for the device configuration request."""

    accountID: str
    token: str
    acceptLanguage: str = DefaultValues.acceptLanguage
    appVersion: str = f'VeSync {DefaultValues.appVersion}'
    debugMode: bool = False
    method: str = 'getAppConfigurationV2'
    phoneBrand: str = DefaultValues.phoneBrand
    phoneOS: str = DefaultValues.phoneOS
    timeZone: str = DefaultValues.timeZone
    traceId: str = field(default_factory=DefaultValues.traceId)
    userCountryCode: str = DefaultValues.userCountryCode
    categories: list[dict[str, str | bool]] = field(
        default_factory=lambda: [
            {
                'category': 'SupportedModelsV3',
                'language': 'en',
                'testMode': False,
                'version': '',
            }
        ]
    )
    recall: bool = False


@dataclass
class ResponseDeviceConfiguration(ResponseCodeModel):
    """Model for the device configuration response.

    Inherits from `BaseResultModel`. The `BaseResultModel` class provides the
    defaults "code" and "msg" fields for the response.

    Attributes:
        result: dict
            The inner model for the 'result' field in the device configuration response.
    """

    result: dict[str, Any]


@dataclass
class ResultDeviceConfiguration(ResponseBaseModel):
    """Model for the device configuration result field.

    This class is referenced by the `ResponseDeviceConfiguration` class.
    """

    configList: list[DeviceConfigurationConfigListItem]


@dataclass
class DeviceConfigurationConfigListItem(ResponseBaseModel):
    """Model for each item in the configList field of the device configuration result.

    This class is referenced by the `ResultDeviceConfiguration` class.
    """

    category: str
    items: list[dict[str, Any]]


@dataclass
class DeviceConfigItem(ResponseBaseModel):
    """Model for each item in the configList field of the device configuration result.

    This class is referenced by the `DeviceConfigurationConfigListItem` class.
    """

    itemKey: str
    itemValue: list[dict[str, Any]]


@dataclass
class DeviceConfigItemValue(ResponseBaseModel):
    """Model for each item in the configList field of the device configuration result.

    This class is referenced by the `DeviceConfigItem` class.
    """

    productLineList: list



"""Classes for VeSync Switch Devices.

This module provides classes for VeSync Switch Devices:

    1. VeSyncSwitch: Abstract Base class for VeSync Switch Devices.
    2. VeSyncWallSwitch: Class for VeSync Wall Switch Devices ESWL01 and ESWL03.
    3. VeSyncDimmerSwitch: Class for VeSync Dimmer Switch Devices ESWD16.


Attributes:
    feature_dict (dict): Dictionary of switch models and their supported features.
        Defines the class to use for each switch model and the list of features
    switch_modules (dict): Dictionary of switch models as keys and their associated
        classes as string values.

Note:
    The switch device is built from the `feature_dict` dictionary and used by the
    `vesync.object_factory` during initial call to pyvesync.vesync.update() and
    determines the class to instantiate for each switch model. These classes should
    not be instantiated manually.
"""

from __future__ import annotations

import logging
from dataclasses import asdict
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices.switch_base import VeSyncSwitch
from pyvesync.const import ConnectionStatus, DeviceStatus
from pyvesync.models import switch_models
from pyvesync.models.bypass_models import RequestBypassV1, TimerModels
from pyvesync.utils.colors import Color
from pyvesync.utils.device_mixins import BypassV1Mixin, process_bypassv1_result
from pyvesync.utils.helpers import Helpers, Timer, Validators

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import SwitchMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel

_LOGGER = logging.getLogger(__name__)


class VeSyncWallSwitch(BypassV1Mixin, VeSyncSwitch):
    """Etekcity standard wall switch.

    Inherits from the BypassV1Mixin and VeSyncSwitch classes.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: SwitchMap
    ) -> None:
        """Initialize Etekcity Wall Switch class.

        Args:
            details (ResponseDeviceDetailsModel): The device details.
            manager (VeSync): The VeSync manager.
            feature_map (SwitchMap): The feature map for the device.
        """
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv1_api(
            RequestBypassV1, method='deviceDetail', endpoint='deviceDetail'
        )

        r = Helpers.process_dev_response(_LOGGER, 'get_details', self, r_dict)
        if r is None:
            return
        resp_model = Helpers.model_maker(
            _LOGGER, switch_models.ResponseSwitchDetails, 'get_details', r, self
        )
        if resp_model is None:
            return
        result = resp_model.result
        if not isinstance(result, switch_models.InternalSwitchResult):
            _LOGGER.warning('Invalid response model for switch details')
            return
        self.state.device_status = result.deviceStatus
        self.state.active_time = result.activeTime
        self.state.connection_status = result.connectionStatus

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        """Toggle switch device."""
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        toggle_str = DeviceStatus.from_bool(toggle)

        r_dict = await self.call_bypassv1_api(
            switch_models.RequestSwitchStatus,
            {'status': toggle_str, 'switchNo': 0},
            'deviceStatus',
            'deviceStatus',
        )

        r = Helpers.process_dev_response(_LOGGER, 'get_details', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> None:
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1GetTimer, method='getTimers', endpoint='timer/getTimers'
        )
        if r_dict is None:
            return
        result_model = process_bypassv1_result(
            self, _LOGGER, 'get_timer', r_dict, TimerModels.ResultV1GetTimer
        )
        if result_model is None:
            return

        timers = result_model.timers
        if not isinstance(timers, list) or len(timers) == 0:
            _LOGGER.debug('No timers found')
            return
        if len(timers) > 1:
            _LOGGER.debug('More than one timer found, using first timer')
        timer = timers[0]
        if not isinstance(timer, TimerModels.TimerItemV1):
            _LOGGER.warning('Invalid timer model')
            return
        self.state.timer = Timer(
            int(timer.counterTimer),
            action=timer.action,
            id=int(timer.timerID),
        )

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            _LOGGER.warning('Invalid action for timer - on/off')
            return False
        update_dict = {
            'action': action,
            'counterTime': str(duration),
        }
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1SetTime,
            update_dict,
            method='addTimer',
            endpoint='timer/addTimer',
        )
        if r_dict is None:
            return False
        result_model = process_bypassv1_result(
            self, _LOGGER, 'set_timer', r_dict, TimerModels.ResultV1SetTimer
        )
        if result_model is None:
            return False
        self.state.timer = Timer(
            int(duration),
            action=action,
            id=int(result_model.timerID),
        )
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            _LOGGER.warning('No timer set, run get_timer() first.')
            return False
        update_dict = {
            'timerId': str(self.state.timer.id),
        }
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1ClearTimer,
            update_dict,
            method='deleteTimer',
            endpoint='timer/deleteTimer',
        )
        if r_dict is None:
            return False
        result = Helpers.process_dev_response(_LOGGER, 'clear_timer', self, r_dict)
        if result is None:
            return False
        self.state.timer = None
        return True


class VeSyncDimmerSwitch(BypassV1Mixin, VeSyncSwitch):
    """Vesync Dimmer Switch Class with RGB Faceplate.

    Inherits from the BypassV1Mixin and VeSyncSwitch classes.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: SwitchMap
    ) -> None:
        """Initialize dimmer switch class.

        Args:
            details (ResponseDeviceDetailsModel): The device details.
            manager (VeSync): The VeSync manager.
            feature_map (SwitchMap): The feature map for the device.
        """
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        r_bytes = await self.call_bypassv1_api(
            switch_models.RequestDimmerDetails,
            method='deviceDetail',
            endpoint='deviceDetail',
        )

        r = Helpers.process_dev_response(_LOGGER, 'get_details', self, r_bytes)
        if r is None:
            return

        resp_model = Helpers.model_maker(
            _LOGGER, switch_models.ResponseSwitchDetails, 'set_timer', r, self
        )
        if resp_model is None:
            return
        result = resp_model.result
        if not isinstance(result, switch_models.InternalDimmerDetailsResult):
            _LOGGER.warning('Invalid response model for dimmer details')
            return
        self.state.active_time = result.activeTime
        self.state.connection_status = result.connectionStatus
        self.state.brightness = result.brightness
        self.state.backlight_status = result.rgbStatus
        new_color = result.rgbValue
        if isinstance(new_color, switch_models.DimmerRGB):
            self.state.backlight_color = Color.from_rgb(
                new_color.red, new_color.green, new_color.blue
            )
        self.state.indicator_status = result.indicatorlightStatus
        self.state.device_status = result.deviceStatus

    @deprecated('switch_toggle() deprecated, use toggle_switch(toggle: bool | None)')
    async def switch_toggle(self, status: str) -> bool:
        """Toggle switch status."""
        return await self.toggle_switch(status == DeviceStatus.ON)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status == 'off'
        toggle_status = DeviceStatus.from_bool(toggle)

        r_bytes = await self.call_bypassv1_api(
            switch_models.RequestDimmerStatus,
            {'status': toggle_status},
            'dimmerPowerSwitchCtl',
            'dimmerPowerSwitchCtl',
        )

        r = Helpers.process_dev_response(_LOGGER, 'toggle_switch', self, r_bytes)
        if r is None:
            return False

        self.state.device_status = toggle_status
        return True

    async def toggle_indicator_light(self, toggle: bool | None = None) -> bool:
        """Toggle indicator light."""
        if toggle is None:
            toggle = self.state.indicator_status == 'off'
        toggle_status = DeviceStatus.from_bool(toggle)

        r_bytes = await self.call_bypassv1_api(
            switch_models.RequestDimmerStatus,
            {'status': toggle_status},
            'dimmerIndicatorLightCtl',
            'dimmerIndicatorLightCtl',
        )

        r = Helpers.process_dev_response(_LOGGER, 'toggle_indicator_light', self, r_bytes)
        if r is None:
            return False

        self.state.indicator_status = toggle_status
        return True

    async def set_backlight_status(
        self,
        status: bool,
        red: int | None = None,
        green: int | None = None,
        blue: int | None = None,
    ) -> bool:
        if red is not None and blue is not None and green is not None:
            new_color = Color.from_rgb(red, green, blue)
        else:
            new_color = None
        status_str = DeviceStatus.from_bool(status)

        update_dict: dict[str, str | dict] = {'status': status_str.value}
        if new_color is not None:
            update_dict['rgbValue'] = asdict(new_color.rgb)
        r_bytes = await self.call_bypassv1_api(
            switch_models.RequestDimmerStatus,
            update_dict,
            'dimmerRgbValueCtl',
            'dimmerRgbValueCtl',
        )

        r = Helpers.process_dev_response(_LOGGER, 'set_rgb_backlight', self, r_bytes)
        if r is None:
            return False

        self.state.backlight_status = status_str
        if new_color is not None:
            self.state.backlight_color = new_color
        self.state.backlight_status = status_str
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_brightness(self, brightness: int) -> bool:
        """Set brightness of dimmer - 1 - 100."""
        if not Validators.validate_zero_to_hundred(brightness):
            _LOGGER.warning('Invalid brightness - must be between 0 and 100')
            return False

        r_bytes = await self.call_bypassv1_api(
            switch_models.RequestDimmerBrightness,
            {'brightness': brightness},
            'dimmerBrightnessCtl',
            'dimmerBrightnessCtl',
        )

        r = Helpers.process_dev_response(_LOGGER, 'get_details', self, r_bytes)
        if r is None:
            return False

        self.state.brightness = brightness
        self.state.device_status = DeviceStatus.ON
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> None:
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1GetTimer, method='getTimers', endpoint='timer/getTimers'
        )
        result_model = process_bypassv1_result(
            self, _LOGGER, 'get_timer', r_dict, TimerModels.ResultV1GetTimer
        )
        if result_model is None:
            return
        timers = result_model.timers
        if not isinstance(timers, list) or len(timers) == 0:
            _LOGGER.info('No timers found')
            return
        if len(timers) > 1:
            _LOGGER.debug('More than one timer found, using first timer')
        timer = timers[0]
        if not isinstance(timer, TimerModels.TimeItemV1):
            _LOGGER.warning('Invalid timer model')
            return
        self.state.timer = Timer(
            int(timer.counterTime),
            action=timer.action,
            id=int(timer.timerID),
        )

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            _LOGGER.warning('Invalid action for timer - on/off')
            return False
        update_dict = {'action': action, 'counterTime': str(duration), 'status': '1'}
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1SetTime,
            update_dict,
            method='addTimer',
            endpoint='timer/addTimer',
        )
        result_model = process_bypassv1_result(
            self, _LOGGER, 'set_timer', r_dict, TimerModels.ResultV1SetTimer
        )
        if result_model is None:
            return False
        self.state.timer = Timer(
            int(duration),
            action=action,
            id=int(result_model.timerID),
        )
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            _LOGGER.debug('No timer set, run get_timer() first.')
            return False
        update_dict = {'timerId': str(self.state.timer.id), 'status': '1'}
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1ClearTimer,
            update_dict,
            method='deleteTimer',
            endpoint='timer/deleteTimer',
        )
        result = Helpers.process_dev_response(_LOGGER, 'clear_timer', self, r_dict)
        if result is None:
            return False
        self.state.timer = None
        return True



"""Test scripts for Etekcity 10A Outlets."""

import pytest
import orjson
from pyvesync.devices.vesyncoutlet import VeSyncOutlet10A
from pyvesync.utils.helpers import Helpers as Helpers
import call_json
import call_json_outlets
from base_test_cases import TestBase
from defaults import TestDefaults

OutletDefaults = call_json_outlets.OutletDefaults

DEV_TYPE_US = 'ESW03-USA'
DEV_TYPE_EU = 'ESW01-EU'
DEV_LIST_DETAIL_EU = call_json.DeviceList.device_list_item(DEV_TYPE_EU)
DEV_LIST_DETAIL_US = call_json.DeviceList.device_list_item(DEV_TYPE_US)
CORRECT_10AUS_LIST = call_json.DeviceList.device_list_response(DEV_TYPE_US)
CORRECT_10AEU_LIST = call_json.DeviceList.device_list_response(DEV_TYPE_EU)

ENERGY_HISTORY = call_json_outlets.ENERGY_HISTORY

CORRECT_10A_DETAILS = call_json_outlets.DETAILS_RESPONSES[DEV_TYPE_US]

BAD_10A_LIST = call_json.DETAILS_BADCODE


class TestVesync10ASwitch(TestBase):
    """Test class for 10A outlets."""

    @pytest.mark.parametrize(
        'mock_return, devtype',
        [(CORRECT_10AEU_LIST, DEV_TYPE_EU), (CORRECT_10AUS_LIST, DEV_TYPE_US)],
    )
    def test_10a_conf(self, mock_return, devtype):
        """Tests that 10A US & EU Outlet is instantiated properly."""
        resp_dict, status = mock_return
        self.mock_api.return_value = resp_dict, status
        self.run_in_loop(self.manager.get_devices)
        outlets = self.manager.outlets
        assert len(outlets) == 1
        outlet = outlets[0]
        assert isinstance(outlet, VeSyncOutlet10A)
        assert outlet.device_name == TestDefaults.name(devtype)
        assert outlet.device_type == devtype
        assert outlet.cid == TestDefaults.cid(devtype)
        assert outlet.uuid == TestDefaults.uuid(devtype)

    def test_10a_details(self):
        """Test 10A get_details()."""
        resp_dict, status = CORRECT_10A_DETAILS
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        outlet = VeSyncOutlet10A(DEV_LIST_DETAIL_US, self.manager)
        self.run_in_loop(outlet.get_details)
        dev_details = outlet.details
        assert outlet.device_status == 'on'
        assert isinstance(dev_details, dict)
        assert dev_details['active_time'] == TestDefaults.active_time
        assert dev_details['energy'] == OutletDefaults.energy
        assert dev_details['power'] == OutletDefaults.power
        assert dev_details['voltage'] == OutletDefaults.voltage
        assert outlet.power == OutletDefaults.power
        assert outlet.voltage == OutletDefaults.voltage

    def test_10a_details_fail(self):
        """Test 10A get_details with Code>0."""
        resp_dict, status = BAD_10A_LIST
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        out = VeSyncOutlet10A(DEV_LIST_DETAIL_EU, self.manager)
        self.run_in_loop(out.get_details)
        assert len(self.caplog.records) == 1
        assert 'details' in self.caplog.text

    def test_10a_onoff(self):
        """Test 10A Device On/Off Methods."""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        out = VeSyncOutlet10A(DEV_LIST_DETAIL_EU, self.manager)
        head = Helpers.req_legacy_headers(self.manager)
        body = Helpers.req_body(self.manager, 'devicestatus')

        body['status'] = 'on'
        body['uuid'] = out.uuid
        on = self.run_in_loop(out.turn_on)
        self.mock_api.assert_called_with(
            '/10a/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert on
        off = self.run_in_loop(out.turn_off)
        body['status'] = 'off'
        self.mock_api.assert_called_with(
            '/10a/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert off

    def test_10a_onoff_fail(self):
        """Test 10A On/Off Fail with Code>0."""
        self.mock_api.return_value = (orjson.dumps({'code': 1}), 400)
        out = VeSyncOutlet10A(DEV_LIST_DETAIL_US, self.manager)
        assert not self.run_in_loop(out.turn_on)
        assert not self.run_in_loop(out.turn_off)

    def test_10a_weekly(self):
        """Test 10A get_weekly_energy."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        out = VeSyncOutlet10A(DEV_LIST_DETAIL_EU, self.manager)
        self.run_in_loop(out.get_weekly_energy)
        body = Helpers.req_body(self.manager, 'energy_week')
        body['uuid'] = out.uuid
        self.mock_api.assert_called_with(
            '/10a/v1/device/energyweek',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = out.energy['week']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert out.weekly_energy_total == 1

    def test_10a_monthly(self):
        """Test 10A get_monthly_energy."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        out = VeSyncOutlet10A(DEV_LIST_DETAIL_EU, self.manager)
        self.run_in_loop(out.get_monthly_energy)
        body = Helpers.req_body(self.manager, 'energy_month')
        body['uuid'] = out.uuid
        self.mock_api.assert_called_with(
            '/10a/v1/device/energymonth',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = out.energy['month']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert out.monthly_energy_total == 1

    def test_10a_yearly(self):
        """Test 10A get_yearly_energy."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        out = VeSyncOutlet10A(DEV_LIST_DETAIL_US, self.manager)
        self.run_in_loop(out.get_yearly_energy)
        body = Helpers.req_body(self.manager, 'energy_year')
        body['uuid'] = out.uuid
        self.mock_api.assert_called_with(
            '/10a/v1/device/energyyear',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = out.energy['year']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert out.yearly_energy_total == 1

    def test_history_fail(self):
        """Test 15A energy failure."""
        bad_history = {'code': 1}
        self.mock_api.return_value = (orjson.dumps(bad_history), 200)
        out = VeSyncOutlet10A(DEV_LIST_DETAIL_US, self.manager)
        self.run_in_loop(out.update_energy)
        assert len(self.caplog.records) == 1
        assert 'weekly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(out.get_monthly_energy)
        assert len(self.caplog.records) == 1
        assert 'monthly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(out.get_yearly_energy)
        assert len(self.caplog.records) == 1
        assert 'yearly' in self.caplog.text



import re
from http import HTTPStatus
from types import TracebackType
from collections import namedtuple
from unittest import mock
from urllib.parse import parse_qs
from yarl import URL
from orjson import dumps, loads
from aiohttp import (
    ClientConnectionError,
    ClientResponseError,
    StreamReader
    )
from multidict import CIMultiDict


RETYPE = type(re.compile(""))

REQUEST_INFO = namedtuple("REQUEST_INFO", ["url", "method", "headers", "real_url"])


def mock_stream(data) -> StreamReader:
    """Mock a stream with data."""
    protocol = mock.Mock(_reading_paused=False)
    stream = StreamReader(protocol, limit=2**16)
    stream.feed_data(data)
    stream.feed_eof()
    return stream


class AiohttpMockSession:
    """MockAiohttpsession."""
    def __init__(self, url, method, response, status, headers=None):
        self.url = URL(url)
        self.method = method
        self.response_bytes = response
        self.status = status
        self._headers = CIMultiDict(headers or {})

    @property
    def headers(self):
        """Return content_type."""
        return self._headers

    async def __aenter__(self):
        return AiohttpClientMockResponse(
            method=self.method,
            url=self.url,
            status=self.status,
            response=self.response_bytes
        )

    async def __aexit__(self, *args):
        pass


class AiohttpClientMockResponse:
    """Mock Aiohttp client response.

    Adapted from home-assistant/core/tests/test_util/aiohttp.py
    """
    def __init__(
        self,
        method,
        url: URL,
        status=HTTPStatus.OK,
        response=None,
        json=None,
        text=None,
        cookies=None,
        exc=None,
        headers=None,
        side_effect=None,
        closing=None,
    ) -> None:
        """Initialize a fake response.

        Args:
            method: str: HTTP method
            url: URL: URL of the response
            status: HTTPStatus: HTTP status code
            response: bytes: Response content
            json: dict: JSON response content
            text: str: Text response content
            cookies: dict: Cookies to set
            exc: Exception: Exception to raise
            headers: dict: Headers to set
            side_effect: Exception: Exception to raise when read
            closing: bool: Simulate closing connection
        """
        if json is not None:
            text = dumps(json)
        if isinstance(text, str):
            response = text.encode("utf-8")
        if response is None:
            response = b""

        self.charset = "utf-8"
        self.method = method
        self._url = url
        self.status = status
        self._response = response
        self.exc = exc
        self.side_effect = side_effect
        self.closing = closing
        self._headers = CIMultiDict(headers or {})
        self._cookies = {}

        if cookies:
            for name, data in cookies.items():
                cookie = mock.MagicMock()
                cookie.value = data
                self._cookies[name] = cookie

    def match_request(self, method, url, params=None):
        """Test if response answers request."""
        if method.lower() != self.method.lower():
            return False

        # regular expression matching
        if isinstance(self._url, RETYPE):
            return self._url.search(str(url)) is not None

        if (
            self._url.scheme != url.scheme
            or self._url.host != url.host
            or self._url.path != url.path
        ):
            return False

        # Ensure all query components in matcher are present in the request
        request_qs = parse_qs(url.query_string)
        matcher_qs = parse_qs(self._url.query_string)
        for key, vals in matcher_qs.items():
            for val in vals:
                try:
                    request_qs.get(key, []).remove(val)
                except ValueError:
                    return False

        return True

    @property
    def request_info(self):
        """Return request info."""
        return REQUEST_INFO(
            url=self.url,
            method=self.method,
            headers=self.headers,
            real_url=self.url)

    @property
    def headers(self):
        """Return content_type."""
        return self._headers

    @property
    def cookies(self):
        """Return dict of cookies."""
        return self._cookies

    @property
    def url(self):
        """Return yarl of URL."""
        return self._url

    @property
    def content_type(self):
        """Return yarl of URL."""
        return self._headers.get("content-type")

    @property
    def content(self):
        """Return content."""
        return mock_stream(self.response)

    async def read(self):
        """Return mock response."""
        return self.response

    async def text(self, encoding="utf-8", errors="strict"):
        """Return mock response as a string."""
        return self.response.decode(encoding, errors=errors)

    async def json(self, encoding="utf-8", content_type=None, loads=loads):
        """Return mock response as a json."""
        return loads(self.response.decode(encoding))

    def release(self):
        """Mock release."""

    def raise_for_status(self):
        """Raise error if status is 400 or higher."""
        if self.status >= 400:
            request_info = mock.Mock(real_url="http://example.com")
            raise ClientResponseError(
                request_info=request_info,
                history=None,  # type: ignore
                status=self.status,
                headers=self.headers,
            )

    def close(self):
        """Mock close."""

    async def wait_for_close(self):
        """Wait until all requests are done.

        Do nothing as we are mocking.
        """

    @property
    def response(self):
        """Property method to expose the response to other read methods."""
        if self.closing:
            raise ClientConnectionError("Connection closed")
        return self._response

    async def __aenter__(self):
        """Enter the context manager."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the context manager."""
